import copy

import nace
from nace.nace_v3 import _plan, _predict_next_world_state
from nace.test_utilities import (convert_focus_set_to_internal_mapping, convert_rules_to_internal_mapping,
                                 get_xy_delta_for_action_list)

def t1_predict():
    world_str_list = [
        'AAAAAAAA..',
        'AcEaaaaa..',
        'Aaaaaaaa..',
        'Aaaabaaa..',
        'Aaaaaaba..', '..........', '..........',
                      '..........', '..........', '..........']
    focus_set = {'A': 0, 'E': 3, 'a': 0, 'b': 0, 'C': 8, 'c': 2, 'H': 4, 'B': 5, 'G': 1}
    rules = [

        # if the agents state is included in the rules, it must match the actual agent state (including score),
        # otherwise the action does not fire.

        # action,agent,   precon_1,      precon_n,       consequences
        #        state                                   world       agent
        #                                                            reward, v0, v1...
        (('down', (0,1),    (1, 0, 'a'),  (0, 0, 'E')),     (0, 0, 'a', (1.0,))),  # new rule
        (('down', (0,1),    (-1, 0, 'E'), (0, 0, 'a')),     (0, 0, 'E', (1.0,))),  # new rule
    ] #            ^- terminted == 0
    #                ^___ arbritary agent value (passed into agent constructor as well)
    world_state_values = {'int_2from_rgb_list': [],
                          'int_2from_char_list': [(-1, 'A'), (-2, '.'), (2617493089, 'B'), (3144150970, 'C'),
                                                  (3712765165, 'D'), (2460067371, 'E'), (961690427, 'F'),
                                                  (1630428625, 'G'), (442688737, 'H'), (1147995044, 'a'),
                                                  (2117949728, 'b'), (1992348315, 'c')], 'raw_cell_shape': (64, 64, 3),
                          'raw_nptype': "<class 'numpy.uint8'>"}
    agent = nace.agent_module.Agent(rc_loc=(1, 2), score=0.0, terminated=0, values_excluding_prefix=[1])


    # hand coded after this line

    observed_times_str_list=[
        '10,10,10,10,10,10,10,10,-inf,-inf',
        '10,12,12,10,10,10,10,10,-inf,-inf',
        '10,10,10,10,10,10,10,10,-inf,-inf',
        '10,10,10,10,10,10,10,10,-inf,-inf',
        '10,10,10,10,10,10,10,10,-inf,-inf',
        '-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf',
        '-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf',
        '-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf',
        '-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf',
        '-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf,-inf',
    ]

    agent_char = world_str_list[agent.get_rc_loc()[0]][agent.get_rc_loc()[1]]

    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        world_str_list,
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=[agent_char], # must match actual agent char
        observed_times_str_list=observed_times_str_list,
        int_2from_char_mapping=world_state_values["int_2from_char_list"]
    )
    world.multiworld_print([{"World": world, "Color": nace.color_codes.color_code_white_on_blue}])

    # convert focus_set and rules to use embeddings rather than char (which is used for convenience)
    new_focus_set = convert_focus_set_to_internal_mapping(focus_set, world)
    new_rules = convert_rules_to_internal_mapping(rules, world)

    plan_actions = ["down"]
    planworld = copy.deepcopy(world)
    planagent = copy.deepcopy(agent)
    for i in range(0, len(plan_actions)):
        planworld, planagent, _, __, agent_values_delta, score_delta = nace.nace_v3._predict_next_world_state(
            new_focus_set,
            copy.deepcopy(planworld), plan_actions[i], new_rules, planagent
        )
    world.multiworld_print([{"Caption": f"Start:",
                             "World": world,
                             "Color": nace.color_codes.color_code_white_on_black},
                            {"Caption": f"At end of plan:\n" + str(
                                nace.prettyprint.prettyprint_all_actions(plan_actions)),
                             "World": planworld,
                             "Color": nace.color_codes.color_code_white_on_red},
                            ]
                           )

    ch = planworld.get_char_at_rc(2,2, agent_indication_embedded_value_list=[])
    print("ch",ch)
    print("agent_values_delta",agent_values_delta)
    assert ch == 'E'
    assert agent_values_delta == [1.0, 0, 0] # score increase by 1 even though 2 rules fired.

    # test part b - two moves down
    print("________________________________")

    world.multiworld_print([{"Caption": f"Start:",
                             "World": world,
                             "Color": nace.color_codes.color_code_white_on_black},
                            ]
                           )
    plan_actions = ["down", "down"]
    planworld = copy.deepcopy(world)
    planagent = copy.deepcopy(agent)
    for i in range(0, len(plan_actions)):
        planworld, planagent, _, __, agent_values_delta, score_delta = nace.nace_v3._predict_next_world_state(
            new_focus_set,
            copy.deepcopy(planworld), plan_actions[i], new_rules, planagent
        )

        world.multiworld_print([
                                {"Caption": f"after move:"+str(i)+"\n" + str(
                                    nace.prettyprint.prettyprint_all_actions(plan_actions)),
                                 "World": planworld,
                                 "Color": nace.color_codes.color_code_white_on_red},
                                ]
                               )

    ch = planworld.get_char_at_rc(3, 2, agent_indication_embedded_value_list=[])
    print("ch",ch)
    assert ch == 'E'
    assert agent_values_delta == [1.0, 0, 0]  # score increase by 1 even though 2 rules fired.

    print()









def t2_check_plan_roll_forward():
    """
    @return:
    """
    print("___________t2_check_plan_roll_forward()________")
    world_str_list = [
        ['oooooooooooo',
         'o   o   x  o',
         'o      f f o',
         'o   oooooooo',
         'o       u  o',
         'o          o',
         'oooooooooooo'], (),
        ['25,25,25,23,24,26,26,26,26,26,26,26,',
         '25,25,25,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,26,26,26,26,26,26,26,',
         '20,21,22,23,24,25,25,25,25,25,25,25,',
         '19,19,19,19,19,19,19,16,15,14,13,12,',
         '18,18,18,18,18,18,18,16,15,14,13,12,']]
    rules = {(('left', (0,), (0, 0, ' '), (0, 1, 'x')), (0, 0, 'x', (0,))),
             (('left', (0,), (0, 0, 'u'), (0, 1, 'x')), (0, 0, 'x', (0,))),
             (('up', (0,), (0, 0, 'o'), (1, 0, 'x')), (0, 0, 'o', (0,))),
             (('down', (0,), (0, 0, 'x'), (1, 0, ' ')), (0, 0, ' ', (0,))),
             (('up', (0,), (0, 0, ' '), (1, 0, 'x')), (0, 0, 'x', (0,))),
             (('up', (0,), (-1, 0, ' '), (0, 0, 'x')), (0, 0, ' ', (0,))),
             (('right', (0,), (0, -1, 'x'), (0, 0, 'o')), (0, 0, 'o', (0,))),
             (('down', (0,), (-1, 0, 'x'), (0, 0, ' ')), (0, 0, 'x', (0,))),
             (('right', (0,), (0, -1, 'x'), (0, 0, 'u')), (0, 0, 'x', (0,))),
             (('down', (0,), (0, 0, 'x'), (1, 0, 'u')), (0, 0, 'u', (0,))),
             (('down', (0,), (-1, 0, 'x'), (0, 0, 'u')), (0, 0, 'x', (0,))),
             (('right', (0,), (0, -1, 'x'), (0, 0, ' ')), (0, 0, 'x', (0,))),
             (('left', (0,), (0, 0, 'o'), (0, 1, 'x')), (0, 0, 'o', (0,))),
             (('down', (0,), (-1, 0, 'x'), (0, 0, 'o')), (0, 0, 'o', (0,))),
             (('right', (0,), (0, 0, 'x'), (0, 1, 'u')), (0, 0, 'u', (0,))),
             (('right', (0,), (0, 0, 'x'), (0, 1, ' ')), (0, 0, ' ', (0,))),
             (('left', (0,), (0, -1, 'u'), (0, 0, 'x')), (0, 0, 'u', (0,))),
             (('left', (0,), (0, -1, ' '), (0, 0, 'x')), (0, 0, ' ', (0,))),
             (('up', (0,), (0, 0, 'u'), (1, 0, 'x')), (0, 0, 'x', (0,))),
             (('up', (0,), (-1, 0, 'u'), (0, 0, 'x')), (0, 0, 'u', (0,))),

             (('left', (0,), (0, 0, 'f'), (0, 1, 'x')), (0, 0, 'x', (1,))),
             (('left', (0,), (0, -1, 'f'), (0, 0, 'x')), (0, 0, ' ', (1,))),
             (('up', (0,), (0, 0, 'f'), (1, 0, 'x')), (0, 0, 'x', (1,))),
             (('up', (0,), (-1, 0, 'f'), (0, 0, 'x')), (0, 0, ' ', (1,))),
             (('down', (0,), (0, 0, 'x'), (1, 0, 'f')), (0, 0, ' ', (1,))),
             (('down', (0,), (-1, 0, 'x'), (0, 0, 'f')), (0, 0, 'x', (1,))),
             (('right', (0,), (0, -1, 'x'), (0, 0, 'f')), (0, 0, 'x', (1,))),
             (('right', (0,), (0, 0, 'x'), (0, 1, 'f')), (0, 0, ' ', (1,))),
             }
    actions = ['left', 'right', 'up', 'down']
    focus_set = {'f': 1, 'u': 1, 'x': 9}
    agent = nace.agent_module.Agent((1, 8), 0, 0, ())

    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        world_str_list[0],
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=['x'],
        observed_times_str_list=world_str_list[2]
    )
    world.multiworld_print([{"World": world, "Color": nace.color_codes.color_code_white_on_blue}])

    # convert focus_set and rules to use embeddings rather than char (which is used for convenience)
    new_focus_set = convert_focus_set_to_internal_mapping(focus_set, world)
    new_rules = convert_rules_to_internal_mapping(rules, world)


    # ======

    # check  _predict_next_world_state() if gived the exact correct moves gives a +ve score
    # yes this should be a seperate test, but the code is changing so fast.

    w_t_plus_1, planagent_1, new_AIRIS_confidence, __age, agent_values_delta, predicted_score_delta = (
        _predict_next_world_state(
            new_focus_set,
            copy.deepcopy(world), 'left', new_rules, agent, custom_goal=None
        ))
    w_t_plus_2, planagent_2,new_AIRIS_confidence, __age, agent_values_delta, predicted_score_delta = (
        _predict_next_world_state(
            new_focus_set,
            copy.deepcopy(w_t_plus_1), 'down', new_rules, agent, custom_goal=None
        ))
    assert predicted_score_delta == 1.0
    assert agent_values_delta[0] == 1.0

    # =========

    (lowest_conf_actions, lowest_AIRIS_confidence, lowest_conf_achieves_goal, lowest_conf_stopping_reason,
     oldest_age_actions, oldest_age, oldest_age_achieves_goal, oldest_age_stopping_reason,
     biggest_predicted_score_delta_actions, biggest_predicted_score_delta,
     biggest_predicted_score_delta_stopping_reason,
     smallest_predicted_score_delta_actions, smallest_predicted_score_delta,
     smallest_predicted_score_delta_stopping_reason,
     debbugging_data) = (
        _plan(
            world,
            new_rules,
            actions,
            new_focus_set,
            agent))

    xy_dt2 = get_xy_delta_for_action_list(biggest_predicted_score_delta_actions)
    assert xy_dt2 == (-1, 1) or xy_dt2 == (1, 1) # go for left food (dependent on action order?)
    assert biggest_predicted_score_delta == 1

    # roll forward the world according to the plan
    planworld = copy.deepcopy(world)
    planagent = copy.deepcopy(agent)
    for i in range(0, len(biggest_predicted_score_delta_actions)):
        planworld, planagent, _, __, ___, ____ = nace.nace_v3._predict_next_world_state(
            new_focus_set,
            copy.deepcopy(planworld), biggest_predicted_score_delta_actions[i], new_rules, planagent
        )

    world.multiworld_print([{"Caption": f"Start:",
                             "World": world,
                             "Color": nace.color_codes.color_code_white_on_black},
                            {"Caption": f"At end of plan:\n" + str(
                                nace.prettyprint.prettyprint_all_actions(biggest_predicted_score_delta_actions)),
                             "World": planworld,
                             "Color": nace.color_codes.color_code_white_on_red},
                            ]
                           )

    counts = planworld.get_board_char_counts(agent_indication_embedded_value_list=[])
    print("counts", counts)
    assert counts['f'] == 1  # the x should be over one of the foods.


def t3_frozen_lake_agent_about_to_die():
    world_str_list = ['AAAAAAAA..', 'Acaaaaaa..', 'AaaaEaaa..', 'Aaaabaaa..', '.aaaaaba..', '.aaabaaa..', '..........',
                      '..........', '..........', '..........']
    time_str_list = []
    focus_set = {'E': 3, 'a': 8, 'I': 6, 'b': 1}
    active_rules = [
        (('right', (0,), (0, 0, 'E'), (0, 1, 'a')), (0, 0, 'c', (0.0, 0))),  # E
        (('down', (0,), (0, 0, 'E'), (1, 0, 'b')), (0, 0, 'a', (-1.0, 1))),
        (('right', (0,), (0, 0, 'I'), (0, 1, 'a')), (0, 0, 'a', (0.0, 0))),  # I
        (('down', (0,), (0, 0, 'I'), (1, 0, 'a')), (0, 0, 'a', (0.0, 0))),
        (('right', (0,), (0, -1, 'E'), (0, 0, 'a')), (0, 0, 'I', (0.0, 0))),  # a
        (('down', (0,), (-1, 0, 'I'), (0, 0, 'a')), (0, 0, 'E', (0.0, 0))),
        (('right', (0,), (0, -1, 'I'), (0, 0, 'a')), (0, 0, 'I', (0.0, 0))),
        (('down', (0,), (-1, 0, 'E'), (0, 0, 'b')), (0, 0, 'F', (-1.0, 1))),  # b
    ]  # subset of rule_evidence
    rule_evidence = {(('right', (0,), (0, 0, 'E'), (0, 1, 'a')), (0, 0, 'c', (0.0, 0))): (2, 0),
                     (('right', (0,), (0, -1, 'E'), (0, 0, 'a')), (0, 0, 'I', (0.0, 0))): (2, 0),
                     (('right', (0,), (0, 0, 'I'), (0, 1, 'a')), (0, 0, 'a', (0.0, 0))): (4, 0),
                     (('right', (0,), (0, -1, 'I'), (0, 0, 'a')), (0, 0, 'I', (0.0, 0))): (4, 0),
                     (('down', (0,), (-1, 0, 'I'), (0, 0, 'a')), (0, 0, 'E', (0.0, 0))): (2, 0),
                     (('down', (0,), (0, 0, 'I'), (1, 0, 'a')), (0, 0, 'a', (0.0, 0))): (2, 0),
                     (('down', (0,), (0, 0, 'E'), (1, 0, 'b')), (0, 0, 'a', (-1.0, 1))): (1, 0),
                     (('down', (0,), (-1, 0, 'E'), (0, 0, 'b')), (0, 0, 'F', (-1.0, 1))): (1, 0)}
    world_state_values = {'int_2from_rgb_list': [],
                          'int_2from_char_list': [(-1, 'A'), (-2, '.'), (2617493089, 'B'), (3144150970, 'C'),
                                                  (3712765165, 'D'), (2460067371, 'E'), (961690427, 'F'),
                                                  (1630428625, 'G'), (2625964021, 'H'), (442688737, 'I'),
                                                  (1147995044, 'a'), (2117949728, 'b'), (1992348315, 'c')],
                          'raw_cell_shape': (64, 64, 3), 'raw_nptype': "<class 'numpy.uint8'>"}
    agent = nace.agent_module.Agent(rc_loc=(2, 4), score=-1.0, terminated=0, values_excluding_prefix=[])
    actions = ['left', 'down', 'right', 'up']
    behavior_returned = 'BABBLE'
    whole_plan_returned = ['left']

    # hand coded code after this line
    """
    (-1, 'A'),    EXTERNAL WALL
    (-2, '.')     UNOBSERVED
    (2617493089, 'B'),  Agent looking left, white background
    (3144150970, 'C'),  Agent looking away (up), white background
    (3712765165, 'D'),  Agent looking left, white background 
    (2460067371, 'E'),  Agent looking at up (down), white background
    (961690427,  'F'),  Agent having fallen through Ice
    (1630428625, 'G'),  Agent, looking right, white background 
    (2625964021, 'H'), Agent on Goal cell, white background
    (442688737, 'I'),  Agent, looking right, white background

    (1147995044, 'a'), a (made from char map)
    (2117949728, 'b'), b (made from char map)
    (1992348315, 'c'), c (made from char map)
    (1868170614, 'd'), d (made from char map)
    """

    agent_char = world_str_list[agent.get_rc_loc()[0]][agent.get_rc_loc()[1]]

    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        world_str_list,
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=[agent_char],  # must match actual agent char
        observed_times_str_list=[],
        int_2from_char_mapping=world_state_values["int_2from_char_list"]
    )
    world.multiworld_print([{"World": world, "Color": nace.color_codes.color_code_white_on_blue}])

    # convert focus_set and rules to use embeddings rather than char (which is used for convenience)
    new_focus_set = convert_focus_set_to_internal_mapping(focus_set, world)
    new_rules = convert_rules_to_internal_mapping(active_rules, world)

    plan_actions = ['down']
    planworld = copy.deepcopy(world)
    planagent = copy.deepcopy(agent)
    for i in range(0, len(plan_actions)):
        planworld, planagent, _, __, ___, ____ = nace.nace_v3._predict_next_world_state(
            new_focus_set,
            copy.deepcopy(planworld), plan_actions[i], new_rules, planagent
        )

    world.multiworld_print([{"Caption": f"Start:",
                             "World": world,
                             "Color": nace.color_codes.color_code_white_on_black},
                            {"Caption": f"At end of plan:\n" + str(
                                nace.prettyprint.prettyprint_all_actions(plan_actions)),
                             "World": planworld,
                             "Color": nace.color_codes.color_code_white_on_red},
                            ]
                           )

    assert planagent.get_terminated() >= 1
    print("INFO test passed, agent is dead")


if __name__ == "__main__":
    t1_predict()
    t2_check_plan_roll_forward()
    t3_frozen_lake_agent_about_to_die()
    print("TESTS COMPLETE")



"""
utilities for rule evidence

"""
import copy
import json
import os.path
import tempfile

import numpy as np

def _convert_tuples_to_lists(obj, index=0, collection_length=1):
    if hasattr(obj, '__iter__') and not isinstance(obj, str):
        result = []
        for i, item in enumerate(obj):
            result.append(_convert_tuples_to_lists(item, i, len(obj)))
        return result
    if index == 2 and collection_length in [3,4] and isinstance(obj, np.int64):
        obj = int(obj)
    return obj

def _convert_lists_to_tuples(obj):
    if hasattr(obj, '__iter__') and isinstance(obj, list):
        result = copy.deepcopy((obj))
        for i, item in enumerate(result):
            result[i] = _convert_lists_to_tuples(item)
        return tuple(result)
    return obj

def _convert_evidence_to_json_serialisable(rule_evidence :dict):
    # strip types, tuple -> list, dict -> list etc
    result = []
    for k ,v in rule_evidence.items():
        converted_k = _convert_tuples_to_lists(k)
        converted_v = _convert_tuples_to_lists(v)
        result.append( [converted_k ,converted_v ])
    return result


def _convert_from_json_to_dict(list_of_list_evidence):
    # add the types back in.
    result = {}
    for l in list_of_list_evidence:
        rule = _convert_lists_to_tuples(l[0])
        evidence = _convert_lists_to_tuples(l[1])

        result[rule] = evidence

    return result


def save_rule_evidence_to_disk(rule_evidence, filename= "rule_evidence.npy"):
    untyped_rule_evidence = _convert_evidence_to_json_serialisable(rule_evidence)
    with open(filename, "w") as f:
        json.dump(untyped_rule_evidence, f, indent=2)

def load_rule_evidence_from_disk(filename= "rule_evidence.npy"):
    rule_evidence = {}
    if os.path.exists(filename):
        with open(filename, "r") as f:
            untyped_rule_evidence = json.load(f)
        rule_evidence = _convert_from_json_to_dict(untyped_rule_evidence)
    return rule_evidence


def t1():
    data = {(('forward', (0, 0, 0, 0, 0, 0, 0), (0, 0, np.int64(2407174942)), (1, -1, np.int64(2407174942)), (2, 0, np.int64(2204181460))), (0, 0, np.int64(1152208772), (0.0, 0, 0, 0, 0, 0, 0, 0))): (1, 2), (('forward', (0, 0, 0, 0, 0, 0, 0), (-1, 1, np.int64(2407174942)), (0, 0, np.int64(2407174942)), (1, 1, np.int64(2204181460))), (0, 0, np.int64(2829951097), (0.0, 0, 0, 0, 0, 0, 0, 0))): (1, 2), (('forward', (0, 0, 0, 0, 0, 0, 0), (-2, 0, np.int64(2407174942)), (-1, -1, np.int64(2407174942)), (0, 0, np.int64(2204181460))), (0, 0, np.int64(1484579719), (0.0, 0, 0, 0, 0, 0, 0, 0))): (1, 2), (('forward', (0, 0, 0, 0, 0, 0, 0), (0, 0, np.int64(2407174942)), (1, -1, np.int64(2407174942)), (2, 0, np.int64(2204181460))), (0, 0, 2407174942, (0.0, 0, 0, 0, 0, 0, 0, 0))): (1, 1), (('forward', (0, 0, 0, 0, 0, 0, 0), (-1, 1, np.int64(2407174942)), (0, 0, np.int64(2407174942)), (1, 1, np.int64(2204181460))), (0, 0, 2407174942, (0.0, 0, 0, 0, 0, 0, 0, 0))): (1, 1), (('forward', (0, 0, 0, 0, 0, 0, 0), (-2, 0, np.int64(2407174942)), (-1, -1, np.int64(2407174942)), (0, 0, np.int64(2204181460))), (0, 0, 2204181460, (0.0, 0, 0, 0, 0, 0, 0, 0))): (1, 1)}
    # check it encodes and decodes correctly
    encoded = _convert_evidence_to_json_serialisable(data)
    unencoded = _convert_from_json_to_dict(encoded)
    assert unencoded == data
    # check it can serialize OK
    t = json.dumps(encoded)
    print(t)


if __name__ == "__main__":
    t1()


import json

import matplotlib.pyplot as plt
import pandas as pd

try:
    import seaborn # pip install opencv-python. Not supported by pypy, and not central, so disabled when using pypy.
    is_seaborn_installed = True
except ImportError as e:
    is_seaborn_installed = False


def _create_plot_collection(to_graph, save_as=None, x_label="", y_label=""):
    if is_seaborn_installed:
        df = pd.DataFrame.from_dict(to_graph)
        plt.clf()
        # create the plot
        fig, ax = plt.subplots()
        fig.set_size_inches(18.5, 10.5)
        handle = seaborn.lineplot(data=df, ax=ax)
        handle.set(xlabel=x_label, ylabel=y_label)
        fig.show()
        if save_as is not None:
            plt.savefig(save_as, dpi=200)


def _produce_behaviour_chart(statistic_series, save_as):
    sub_sampled = []
    bucket_size = 100
    for i in range(len(statistic_series)):
        if i > bucket_size:
            n = float(bucket_size)
            window_data = statistic_series[i - bucket_size:i]
        else:
            n = float(i + 1)
            window_data = statistic_series[:i]

        sum_curious = sum(r['plan.behavior.curious'] for r in window_data)
        sum_explore = sum(r['plan.behavior.explore'] for r in window_data)
        sum_babble = sum(r['plan.behavior.babble'] for r in window_data)
        sum_achieve = sum(r['plan.behavior.achieve'] for r in window_data)

        sub_sampled.append(
            {
                "curious": sum_curious / n,
                "explore": sum_explore / n,
                "babble": sum_babble / n,
                "achieve": sum_achieve / n
            }
        )
    y_label = "Proportion"
    x_label = "Step"
    _create_plot_collection(sub_sampled, save_as, x_label, y_label)


def _produce_confidence_chart(statistic_series, save_as):
    sub_sampled = []
    for i in range(len(statistic_series)):
        sub_sampled.append(
            {
                "AIRIS_confidence": statistic_series[i]['plan.lowest_AIRIS_confidence'] * 100.0,
            }
        )
    y_label = "AIRIS Confidence"
    x_label = "Step"
    _create_plot_collection(sub_sampled, save_as, x_label, y_label)


def _produce_average_return_chart(statistic_series, save_as):
    sub_sampled = []
    for i in range(len(statistic_series)):
        produce_data_for_index = -1
        if (i == len(statistic_series) - 1):
            produce_data_for_index = i
        if i > 0 and 'episode_count' in statistic_series[i] and statistic_series[i - 1]['episode_count'] != \
                statistic_series[i]['episode_count']:
            produce_data_for_index = i - 1

        if produce_data_for_index > -1:
            reward = statistic_series[produce_data_for_index]['cumulative_gym_reward'] if 'cumulative_gym_reward' in \
                                                                                          statistic_series[
                                                                                              produce_data_for_index] else 0.0
            episode_count = statistic_series[produce_data_for_index]['episode_count'] + 1 if 'episode_count' in \
                                                                                             statistic_series[
                                                                                                 produce_data_for_index] else 1  # zero based
            aer = reward / episode_count
            sub_sampled.append(
                {
                    "average_episode_return": aer
                }
            )
    y_label = "Average return per episode"
    x_label = "Episode"
    _create_plot_collection(sub_sampled, save_as, x_label, y_label)


def produce_charts(statistic_series, prefix):
    _produce_confidence_chart(statistic_series, save_as="./data/" + prefix + "_confidence.png")
    _produce_behaviour_chart(statistic_series, save_as="./data/" + prefix + "_behaviour.png")
    _produce_average_return_chart(statistic_series, save_as="./data/" + prefix + "_average_return.png")


if __name__ == "__main__":
    fn = "../data/" + "series_statistics.json"
    with open(fn, "r") as f:
        statistic_series = json.load(f)
    prefix = ""
    # note the different assumed working directory
    _produce_confidence_chart(statistic_series, save_as="../data/" + prefix + "_confidence.png")
    _produce_behaviour_chart(statistic_series, save_as="../data/" + prefix + "_behaviour.png")
    _produce_average_return_chart(statistic_series, save_as="../data/" + prefix + "_average_return.png")



from nace.hypothesis import Hypothesis_TruthValue
from nace.prettyprint import prettyprint_rule

def t1_pp_rule():
    rule_evidence = {(('right', (0,), (0, 0, 'x'), (0, 1, 'o')), (0, 0, ' ', (-1, 0))): (1, 0)}
    rule = (('right', (0,), (0, 0, 'x'), (0, 1, 'o')), (0, 0, ' ', (-1, 0)))
    line = prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule)
    print(line)
    assert line == "<(AgentStatePreCon=[0], rc[ 0, 0 ]=x, rc[ 0, 1 ]=o, right)=/> (AgentStatePreCon=[0], rc[ 0, 0 ]= , score+=-1)>. TruthValue=(1.0, 0.5)"

if __name__ == "__main__":
    t1_pp_rule()




import copy
import random

import nace

from nace.nace_v3 import _predict_next_world_state, _refine_rules_from_changesets, _build_change_sets, _observe_v2
from nace.test_utilities import convert_focus_set_to_internal_mapping, convert_rules_to_internal_mapping

from nace.agent_module import Agent
from nace.test_utilities import (convert_focus_set_to_internal_mapping, convert_rules_to_internal_mapping,
                                 convert_rule_evidence_to_internal_mapping,
                                 convert_rules_to_char_or_int_mappings)

def place_agent_and_object(input_world_str_list, input_agent, desired_agent_rc_loc, desired_object_rc_loc, object_char, agent_char, int_2from_char_mapping):
    world_str_list = copy.deepcopy(input_world_str_list)
    world_str_list[desired_agent_rc_loc[0]] = world_str_list[desired_agent_rc_loc[0]][:desired_agent_rc_loc[1]] + agent_char + world_str_list[desired_agent_rc_loc[0]][desired_agent_rc_loc[1] + 1:]
    world_str_list[desired_object_rc_loc[0]] = world_str_list[desired_object_rc_loc[0]][:desired_object_rc_loc[1]] + object_char + world_str_list[desired_object_rc_loc[0]][desired_object_rc_loc[1] + 1:]
    world, _ = nace.world_module_numpy.NPWorld.from_string(
        world_str_list,
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=[agent_char], # must match actual agent char
        observed_times_str_list=[],
        int_2from_char_mapping=int_2from_char_mapping
    )
    agent = Agent(desired_agent_rc_loc, input_agent.get_score(), input_agent.get_terminated(), input_agent.get_values_exc_prefix())
    return world, agent


def t3_cup_on_table():
    """
    Test we add rules that add +1 to score for putting the cup on the table.
    @return:
    """

    original_focus_set = {'T': 0, 'u': 9, 'x': 65}
    original_rule_evidence = {(('left', (0,), (0, 0, ' '), (0, 1, 'x')), (0, 0, 'x', (0, 0))): (1, 0), (('up', (0,), (0, 0, ' '), (1, 0, 'x')), (0, 0, 'x', (0, 0))): (1, 0), (('right', (0,), (0, -1, 'x'), (0, 0, ' ')), (0, 0, 'x', (0, 0))): (1, 0), (('down', (0,), (-1, 0, 'x'), (0, 0, ' ')), (0, 0, 'x', (0, 0))): (1, 0), (('left', (0,), (0, -1, ' '), (0, 0, 'x')), (0, 0, ' ', (0, 0))): (1, 0), (('up', (0,), (-1, 0, ' '), (0, 0, 'x')), (0, 0, ' ', (0, 0))): (1, 0), (('right', (0,), (0, 0, 'x'), (0, 1, ' ')), (0, 0, ' ', (0, 0))): (1, 0), (('down', (0,), (0, 0, 'x'), (1, 0, ' ')), (0, 0, ' ', (0, 0))): (1, 0), (('left', (0,), (0, 0, 'o'), (0, 1, 'x')), (0, 0, 'o', (0, 0))): (1, 0), (('up', (0,), (0, 0, 'o'), (1, 0, 'x')), (0, 0, 'o', (0, 0))): (1, 0), (('right', (0,), (0, -1, 'x'), (0, 0, 'o')), (0, 0, 'o', (0, 0))): (1, 0), (('down', (0,), (-1, 0, 'x'), (0, 0, 'o')), (0, 0, 'o', (0, 0))): (1, 0), (('down', (0,), (-1, 0, 'x'), (0, 0, 'u')), (0, 0, 'x', (0, 0))): (1, 0), (('left', (0,), (0, 0, 'u'), (0, 1, 'x')), (0, 0, 'x', (0, 0))): (1, 0), (('up', (0,), (0, 0, 'u'), (1, 0, 'x')), (0, 0, 'x', (0, 0))): (1, 0), (('right', (0,), (0, -1, 'x'), (0, 0, 'u')), (0, 0, 'x', (0, 0))): (2, 0), (('down', (0,), (0, 0, 'x'), (1, 0, 'T')), (0, 0, 'x', (0, 0))): (1, 0), (('left', (0,), (0, -1, 'T'), (0, 0, 'x')), (0, 0, 'x', (0, 0))): (1, 0), (('up', (0,), (-1, 0, 'T'), (0, 0, 'x')), (0, 0, 'x', (0, 0))): (1, 0), (('right', (0,), (0, 0, 'x'), (0, 1, 'T')), (0, 0, 'x', (0, 0))): (1, 0), (('down', (0,), (-1, 0, 'x'), (0, 0, 'T')), (0, 0, 'T', (0, 0))): (1, 0), (('left', (0,), (0, 0, 'T'), (0, 1, 'x')), (0, 0, 'T', (0, 0))): (1, 0), (('up', (0,), (0, 0, 'T'), (1, 0, 'x')), (0, 0, 'T', (0, 0))): (1, 0), (('right', (0,), (0, -1, 'x'), (0, 0, 'T')), (0, 0, 'T', (0, 0))): (1, 0), (('right', (0,), (0, 0, 'x'), (0, 1, 'u')), (0, 0, 'u', (0, 0))): (1, 0), (('down', (0,), (0, 0, 'x'), (1, 0, 'u')), (0, 0, 'u', (0, 0))): (1, 0), (('left', (0,), (0, -1, 'u'), (0, 0, 'x')), (0, 0, 'u', (0, 0))): (1, 0), (('up', (0,), (-1, 0, 'u'), (0, 0, 'x')), (0, 0, 'u', (0, 0))): (1, 0)}
    original_rules = {(('left', (0,), (0, -1, 'u'), (0, 0, 'x')), (0, 0, 'u', (0, 0))), (('up', (0,), (0, 0, 'T'), (1, 0, 'x')), (0, 0, 'T', (0, 0))), (('right', (0,), (0, -1, 'x'), (0, 0, 'u')), (0, 0, 'x', (0, 0))), (('right', (0,), (0, -1, 'x'), (0, 0, ' ')), (0, 0, 'x', (0, 0))), (('left', (0,), (0, 0, 'T'), (0, 1, 'x')), (0, 0, 'T', (0, 0))), (('up', (0,), (0, 0, ' '), (1, 0, 'x')), (0, 0, 'x', (0, 0))), (('up', (0,), (0, 0, 'o'), (1, 0, 'x')), (0, 0, 'o', (0, 0))), (('right', (0,), (0, 0, 'x'), (0, 1, ' ')), (0, 0, ' ', (0, 0))), (('left', (0,), (0, 0, 'o'), (0, 1, 'x')), (0, 0, 'o', (0, 0))), (('left', (0,), (0, 0, ' '), (0, 1, 'x')), (0, 0, 'x', (0, 0))), (('right', (0,), (0, 0, 'x'), (0, 1, 'T')), (0, 0, 'x', (0, 0))), (('left', (0,), (0, -1, ' '), (0, 0, 'x')), (0, 0, ' ', (0, 0))), (('down', (0,), (0, 0, 'x'), (1, 0, ' ')), (0, 0, ' ', (0, 0))), (('left', (0,), (0, -1, 'T'), (0, 0, 'x')), (0, 0, 'x', (0, 0))), (('right', (0,), (0, 0, 'x'), (0, 1, 'u')), (0, 0, 'u', (0, 0))), (('down', (0,), (0, 0, 'x'), (1, 0, 'u')), (0, 0, 'u', (0, 0))), (('up', (0,), (-1, 0, 'u'), (0, 0, 'x')), (0, 0, 'u', (0, 0))), (('up', (0,), (-1, 0, 'T'), (0, 0, 'x')), (0, 0, 'x', (0, 0))), (('down', (0,), (-1, 0, 'x'), (0, 0, 'T')), (0, 0, 'T', (0, 0))), (('down', (0,), (0, 0, 'x'), (1, 0, 'T')), (0, 0, 'x', (0, 0))), (('right', (0,), (0, -1, 'x'), (0, 0, 'T')), (0, 0, 'T', (0, 0))), (('left', (0,), (0, 0, 'u'), (0, 1, 'x')), (0, 0, 'x', (0, 0))), (('down', (0,), (-1, 0, 'x'), (0, 0, ' ')), (0, 0, 'x', (0, 0))), (('right', (0,), (0, -1, 'x'), (0, 0, 'o')), (0, 0, 'o', (0, 0))), (('down', (0,), (-1, 0, 'x'), (0, 0, 'u')), (0, 0, 'x', (0, 0))), (('up', (0,), (-1, 0, ' '), (0, 0, 'x')), (0, 0, ' ', (0, 0))), (('down', (0,), (-1, 0, 'x'), (0, 0, 'o')), (0, 0, 'o', (0, 0))), (('up', (0,), (0, 0, 'u'), (1, 0, 'x')), (0, 0, 'x', (0, 0)))}

    original_agent = nace.agent_module.Agent(rc_loc=[2, 10], score=0, terminated=0, values_excluding_prefix=[0])
    new_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=1, terminated=0, values_excluding_prefix=[0])
    predicted_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=1, terminated=0, values_excluding_prefix=[0])
    action = 'left'
    original_negrules = set()

    oldworld_string_list = (
        ['oooooooooooo','o          o','o        uxo','o     ooooTo','o          o','o          o','oooooooooooo'],
        (0, 0), ['38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,64,67,67,67,67,67','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])
    oldworld, old_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        oldworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2]
    )

    newworld_string_list = (
        ['oooooooooooo','o          o','o        xuo','o     ooooTo','o          o','o          o','oooooooooooo'],
        (0,0), ['38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,64,68,68,68,68,68','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])
    newworld, new_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        newworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2])


    encoded_focus_set = convert_focus_set_to_internal_mapping(original_focus_set, oldworld)
    encoded_rules = convert_rules_to_internal_mapping(original_rules, oldworld)
    encoded_negrules = convert_rules_to_internal_mapping(original_negrules, oldworld)



    # Call observe in order to learn dynamics
    calculated_focus_set, calculated_rule_evidence, calculated_new_rules, calculated_new_negrules = _observe_v2(
        encoded_focus_set, original_rule_evidence, oldworld, action, newworld, encoded_rules, encoded_negrules,
        newworld, pre_action_agent=original_agent, ground_truth_post_action_agent=new_agent,
        predicted_agent=predicted_agent,
        unobserved_code='.',
        object_count_threshold=1
    )

    print("New Rules:")
    rules_that_increase_score_count = 0
    for rule in calculated_new_rules:
        prediction_state_value_deltas = rule[1]
        state_value_deltas = prediction_state_value_deltas[-1]
        # print('prediction_state_value_deltas', prediction_state_value_deltas)
        if state_value_deltas[0] > 0:
            rules_that_increase_score_count += 1
            line = nace.prettyprint.prettyprint_rule(rule_evidence=calculated_rule_evidence,
                                                     Hypothesis_TruthValue=nace.hypothesis.Hypothesis_TruthValue,
                                                     rule=rule
                                                     )
            print(line)

    print("INFO rules_that_increase_score_count", rules_that_increase_score_count)
    assert rules_that_increase_score_count == 2 # was 8, will be 2 if we do not assume eucldian rules.
    # ideally we would rebuild or edit all the rules to include the terminated flag in precondition. # expect 2 rules for each action (1 for source cell, 1 for dest) * 4 actions == 8 new rules



def t1_simple_move_left():
    """
    Move left, check new rules are derived.

    @return:
    """

    original_focus_set = { 'x': 65}
    original_rule_evidence = {}
    original_rules = set()

    old_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=0, terminated=0, values_excluding_prefix=[0])
    new_agent = nace.agent_module.Agent(rc_loc=[2, 8], score=0, terminated=0, values_excluding_prefix=[0])
    action = 'left'
    original_negrules = set()

    oldworld_string_list = (
        ['oooooooooooo','o          o','o        x o','o          o','o          o','o          o','oooooooooooo'],
        (0, 0), ['38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,64,67,67,67,67,67','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])

    oldworld, old_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        oldworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2]
    )

    newworld_string_list = (
        ['oooooooooooo','o          o','o       x  o','o          o','o          o','o          o','oooooooooooo'],
        (0,0), ['38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,64,68,68,68,68,68','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])
    new_world, new_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        newworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=newworld_string_list[2]
    )

    encoded_focus_set = convert_focus_set_to_internal_mapping(original_focus_set, oldworld)
    encoded_rules = convert_rules_to_internal_mapping(original_rules, oldworld)
    encoded_negrules = convert_rules_to_internal_mapping(original_negrules, oldworld)
    encoded_rule_evidence = original_rule_evidence
    predicted_world, predicted_agent, _, __, values, predicted_score_delta = _predict_next_world_state(
        encoded_focus_set,
        copy.deepcopy(oldworld), action, encoded_rules, old_agent
    )

    new_world.multiworld_print([
        {"World": oldworld,       "Caption":"Old",  "Color": nace.color_codes.color_code_white_on_blue},
        {"World": predicted_world, "Caption":"Pred\nNo Rule, Empty Pred", "Color": nace.color_codes.color_code_white_on_blue},
        {"World": new_world,       "Caption":"New",  "Color": nace.color_codes.color_code_white_on_blue},
    ])

    # Call the function
    calculated_focus_set, calculated_rule_evidence, calculated_new_rules, calculated_new_negrules = _observe_v2(
        encoded_focus_set, encoded_rule_evidence, oldworld, action, new_world, encoded_rules, encoded_negrules,
        predicted_world, pre_action_agent=old_agent, ground_truth_post_action_agent=new_agent,
        predicted_agent=predicted_agent,
        unobserved_code='.',
        object_count_threshold=1
    )

    print("About to check the rules with the action 'left', We expect 8 rules in total, but only if rules are expanded for eurclidean space, 2 rules where the action is left, One for where the agent was, one for where the agent goes to.")
    print("calculated_new_rules",calculated_new_rules)
    assert len(calculated_new_rules) == 2
    left_rule_count = 0
    for rule in calculated_new_rules:
        if rule[0][0] == action:
            nace.prettyprint.prettyprint_rule(rule_evidence=calculated_rule_evidence,
                                              Hypothesis_TruthValue=nace.hypothesis.Hypothesis_TruthValue,
                                              rule=rule
                                              )
            prediction_state_value_deltas = rule[1]
            state_value_deltas = prediction_state_value_deltas[-1]
            left_rule_count += 1
    assert left_rule_count == 2 # we expect 2 rules, i.e. 2 cell values change


def t2_simple_move_left_get_moved_left_and_down():
    """
    Move left, check new rules are derived.

    @return:
    """


    original_focus_set = {}
    original_rule_evidence = {}
    oldrules = set()

    old_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=0, terminated=0, values_excluding_prefix=[0])
    new_agent = nace.agent_module.Agent(rc_loc=[3, 8], score=0, terminated=0, values_excluding_prefix=[0])
    action = 'left'
    oldnegrules = set()

    oldworld_string_list = (
        ['oooooooooooo','o          o','o        x o','o          o','o          o','o          o','oooooooooooo'],
        (0, 0), ['38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,64,67,67,67,67,67','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])
    old_world, _2 = nace.world_module_numpy.NPWorld.from_string(
        oldworld_string_list[0],
        view_dist_x=30,
        view_dist_y=20,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2]
    )

    newworld_string_list = (
        ['oooooooooooo','o          o','o          o','o       x  o','o          o','o          o','oooooooooooo'],
        (0,0), ['38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,64,68,68,68,68,68','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])
    new_world, _1 = nace.world_module_numpy.NPWorld.from_string(
        newworld_string_list[0],
        view_dist_x=30,
        view_dist_y=20,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2]
    )

    predicted_world, predicted_agent, _, __, values, predicted_score_delta = _predict_next_world_state(
        original_focus_set,
        copy.deepcopy(old_world), action, oldrules, old_agent
    )

    new_world.multiworld_print([
        {"World": old_world,       "Caption":"Old",  "Color": nace.color_codes.color_code_white_on_blue},
        {"World": predicted_world, "Caption":"Pred", "Color": nace.color_codes.color_code_white_on_blue},
        {"World": new_world,       "Caption":"New",  "Color": nace.color_codes.color_code_white_on_blue},
    ])


    # # call the split method
    # calculated_focus_set_v2, calculated_rule_evidence_v2, calculated_new_rules_v2, calculated_new_negrules_v2 = _observe_v2(
    #     copy.deepcopy(original_focus_set), copy.deepcopy(original_rule_evidence), old_world, action, new_world, copy.deepcopy(oldrules), copy.deepcopy(oldnegrules),
    #     predicted_world, pre_action_agent=old_agent, ground_truth_post_action_agent=new_agent, unobserved_code='.',
    #     object_count_threshold=1
    # )
    #
    # #

    # # # call the function in 2 parts.
    new_focus_set, adjacent_change_sets, changeset0len = _build_change_sets(
        original_focus_set,
        old_world,
        action,
        new_world,
        predicted_world,
    )
    assert len(adjacent_change_sets) == 1
    assert adjacent_change_sets[0] == {(2,8), (2, 9), (3, 8)} # ideally would only be 2 cells (2, 9), (3, 8)

    adjacent_change_sets = [{(2, 9), (3, 8)}]
    changeset0len = 2

    calculated_focus_set2, calculated_rule_evidence2, calculated_new_rules2, calculated_new_negrules2 = _refine_rules_from_changesets(original_focus_set,
                                  original_rule_evidence,
                                  old_world,
                                  action,
                                  new_world,
                                  oldrules,
                                  old_agent,
                                  new_agent,
                                  predicted_agent,
                                  adjacent_change_sets,
                                  copy.deepcopy(oldnegrules),
                                  changeset0len
    )

    print("About to check the rules with the action 'left', We expect 2/8 rules to be left.")
    left_rule_count = 0
    for rule in calculated_new_rules2:
        if rule[0][0] == action:
            line = nace.prettyprint.prettyprint_rule(rule_evidence=calculated_rule_evidence2,
                                                     Hypothesis_TruthValue=nace.hypothesis.Hypothesis_TruthValue,
                                                     rule=rule
                                                     )
            print(line)
            prediction_state_value_deltas = rule[1]
            state_value_deltas = prediction_state_value_deltas[-1]
            left_rule_count += 1
    assert left_rule_count == 2 # we expect 2 rules, i.e. 2 cell values change




def t4_learn_to_drop_agent_preconditions():
    """
    Move left, check new rules are derived.

    @return:
    """
    # nace.hypothesis.Hypothesis_DisableLRUDMovementOpAssumptions()
    # nace.hypothesis.Hypothesis_UseLRUDMovementOpAssumptions(
    #     'left',
    #     'right',
    #     'up',
    #     'down',
    #     nace.world_module.drop,
    #     DisableLRUDOpSymmetryAssumptionFlag=False,
    # )

    original_focus_set = { 'x': 65}
    original_rule_evidence = {}
    original_rules = set()

    old_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=0, terminated=0, values_excluding_prefix=[0])
    new_agent = nace.agent_module.Agent(rc_loc=[2, 8], score=0, terminated=0, values_excluding_prefix=[0])
    action = 'left'
    original_negrules = set()

    oldworld_string_list = (
        ['oooooooooooo','o          o','o        x o','o          o','o          o','o          o','oooooooooooo'],
        (0, 0), ['38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,65,67,67,67,67,67','38,44,49,54,59,64,64,67,67,67,67,67','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])

    oldworld, old_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        oldworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2]
    )

    newworld_string_list = (
        ['oooooooooooo','o          o','o       x  o','o          o','o          o','o          o','oooooooooooo'],
        (0,0), ['38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,65,68,68,68,68,68','38,44,49,54,59,64,64,68,68,68,68,68','34,34,34,34,34,34,34,31,30,29,28,27','33,33,33,33,33,33,33,31,30,29,28,27'])
    new_world, new_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        newworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=newworld_string_list[2]
    )

    encoded_focus_set = convert_focus_set_to_internal_mapping(original_focus_set, oldworld)
    encoded_rules = convert_rules_to_internal_mapping(original_rules, oldworld)
    encoded_negrules = convert_rules_to_internal_mapping(original_negrules, oldworld)
    encoded_rule_evidence = original_rule_evidence
    predicted_world, predicted_agent, _, __, values, predicted_score_delta = _predict_next_world_state(
        encoded_focus_set,
        copy.deepcopy(oldworld), action, encoded_rules, old_agent
    )

    new_world.multiworld_print([
        {"World": oldworld,       "Caption":"Old",  "Color": nace.color_codes.color_code_white_on_blue},
        {"World": predicted_world, "Caption":"Pred\nNo Rule, Empty Pred", "Color": nace.color_codes.color_code_white_on_blue},
        {"World": new_world,       "Caption":"New",  "Color": nace.color_codes.color_code_white_on_blue},
    ])

    # copy values to check are not being mutataed.
    backup_encoded_focus_set = copy.deepcopy(encoded_focus_set)
    backup_encoded_rule_evidence = copy.deepcopy(encoded_rule_evidence)
    backup_oldworld = copy.deepcopy(oldworld)
    backup_new_world = copy.deepcopy(new_world)
    backup_encoded_rules = copy.deepcopy(encoded_rules)
    backup_encoded_negrules = copy.deepcopy(encoded_negrules)
    backup_predicted_world = copy.deepcopy(predicted_world)
    backup_old_agent = copy.deepcopy(old_agent)
    backup_new_agent = copy.deepcopy(new_agent)

    # Call the function
    calculated_focus_set, calculated_rule_evidence, calculated_new_rules, calculated_new_negrules = _observe_v2(
        encoded_focus_set, encoded_rule_evidence, oldworld, action, new_world, encoded_rules, encoded_negrules,
        predicted_world, pre_action_agent=old_agent, ground_truth_post_action_agent=new_agent,
        predicted_agent=predicted_agent,
        unobserved_code='.',
        object_count_threshold=1
    )

    # check to see we mutataed nothing
    assert backup_encoded_focus_set == encoded_focus_set
    assert backup_encoded_rule_evidence == encoded_rule_evidence
    assert backup_oldworld == oldworld
    assert backup_new_world == new_world
    assert backup_encoded_rules == encoded_rules
    assert backup_encoded_negrules == encoded_negrules
    assert backup_predicted_world == predicted_world
    assert backup_old_agent == old_agent
    assert backup_new_agent == new_agent

    print("About to check the rules with the action 'left', We expect 2 rules in total as there was no movement assumptions")
    left_rule_count = 0
    for rule in calculated_new_rules:
        if rule[0][0] == action:
            line = nace.prettyprint.prettyprint_rule(
                  rule_evidence=calculated_rule_evidence,
                  Hypothesis_TruthValue=nace.hypothesis.Hypothesis_TruthValue,
                  rule=rule
                  )
            print(line)
            prediction_state_value_deltas = rule[1]
            state_value_deltas = prediction_state_value_deltas[-1]
            left_rule_count += 1
    assert left_rule_count == 2 # we expect 2 rules, i.e. 2 cell values change

    # perturb the agent so that agent precondition(s) don't match current rules and call _observe() (again),
    # and check rules were updated in the correct way
    print("INFO -------- new test start ------")
    perturbed_agent = copy.deepcopy(new_agent)
    perturbed_agent.set_values_inc_prefix((1, 1)) # score += 1, state[0] += 1

    calculated_focus_set_2, calculated_rule_evidence_2, calculated_new_rules_2, calculated_new_negrules_2 = _observe_v2(
        encoded_focus_set, encoded_rule_evidence, oldworld, action, new_world, encoded_rules, encoded_negrules,
        predicted_world, pre_action_agent=old_agent, ground_truth_post_action_agent=perturbed_agent,
        predicted_agent=predicted_agent,
        unobserved_code='.',
        object_count_threshold=1
    )

    # assert score and state0 of agent rules were both 0 in original rules.
    for rule in calculated_new_rules:
        assert rule[1][3][0] == 0
        assert rule[1][3][1] == 0

    # assert score and state0 of agent rules are now both 1 in both rules.
    for rule in calculated_new_rules_2:
        assert rule[1][3][0] == 1
        assert rule[1][3][1] == 1
    print("test passed")


def t5_learn_agent_status_change_key_pickup():
    """
    pick up a key, learn agent state changes.

    @return:
    """
    # nace.hypothesis.Hypothesis_DisableLRUDMovementOpAssumptions()
    # nace.hypothesis.Hypothesis_UseLRUDMovementOpAssumptions(
    #     'left',
    #     'right',
    #     'up',
    #     'down',
    #     nace.world_module.drop,
    #     DisableLRUDOpSymmetryAssumptionFlag=False,
    # )

    original_focus_set = {'x': 65, 'k':1}
    original_rule_evidence = {}
    original_rules = set()

    old_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=0, terminated=0, values_excluding_prefix=[0])
    new_agent = nace.agent_module.Agent(rc_loc=[2, 9], score=0, terminated=0, values_excluding_prefix=[1])
    action = "pickUpKey"
    original_negrules = set()

    oldworld_string_list = (
        ['oooooooooooo', 'o          o', 'o        xko', 'o          o', 'o          o', 'o          o',
         'oooooooooooo'],
        (0, 0), ['38,44,49,54,59,64,65,67,67,67,67,67', '38,44,49,54,59,64,65,67,67,67,67,67',
                 '38,44,49,54,59,64,65,67,67,67,67,67', '38,44,49,54,59,64,65,67,67,67,67,67',
                 '38,44,49,54,59,64,64,67,67,67,67,67', '34,34,34,34,34,34,34,31,30,29,28,27',
                 '33,33,33,33,33,33,33,31,30,29,28,27'])

    oldworld, old_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        oldworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=oldworld_string_list[2]
    )

    newworld_string_list = (
        ['oooooooooooo', 'o          o', 'o        x o', 'o          o', 'o          o', 'o          o',
         'oooooooooooo'],
        (0, 0), ['38,44,49,54,59,64,65,68,68,68,68,68', '38,44,49,54,59,64,65,68,68,68,68,68',
                 '38,44,49,54,59,64,65,68,68,68,68,68', '38,44,49,54,59,64,65,68,68,68,68,68',
                 '38,44,49,54,59,64,64,68,68,68,68,68', '34,34,34,34,34,34,34,31,30,29,28,27',
                 '33,33,33,33,33,33,33,31,30,29,28,27'])
    new_world, new_agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        newworld_string_list[0],
        view_dist_x=3,
        view_dist_y=2,
        agent_char_code_list=['x'],
        observed_times_str_list=newworld_string_list[2]
    )

    encoded_focus_set = convert_focus_set_to_internal_mapping(original_focus_set, oldworld)
    encoded_rules = convert_rules_to_internal_mapping(original_rules, oldworld)
    encoded_negrules = convert_rules_to_internal_mapping(original_negrules, oldworld)
    encoded_rule_evidence = original_rule_evidence
    predicted_world, predicted_agent, _, __, values, predicted_score_delta = _predict_next_world_state(
        encoded_focus_set,
        copy.deepcopy(oldworld), action, encoded_rules, old_agent
    )

    new_world.multiworld_print([
        {"World": oldworld, "Caption": "Old", "Color": nace.color_codes.color_code_white_on_blue},
        {"World": predicted_world, "Caption": "Pred\nNo Rule, Empty Pred",
         "Color": nace.color_codes.color_code_white_on_blue},
        {"World": new_world, "Caption": "New", "Color": nace.color_codes.color_code_white_on_blue},
    ])

    # copy values to check are not being mutataed.
    backup_encoded_focus_set = copy.deepcopy(encoded_focus_set)
    backup_encoded_rule_evidence = copy.deepcopy(encoded_rule_evidence)
    backup_oldworld = copy.deepcopy(oldworld)
    backup_new_world = copy.deepcopy(new_world)
    backup_encoded_rules = copy.deepcopy(encoded_rules)
    backup_encoded_negrules = copy.deepcopy(encoded_negrules)
    backup_predicted_world = copy.deepcopy(predicted_world)
    backup_old_agent = copy.deepcopy(old_agent)
    backup_new_agent = copy.deepcopy(new_agent)

    # Call the function
    calculated_focus_set, calculated_rule_evidence, calculated_new_rules, calculated_new_negrules = _observe_v2(
        encoded_focus_set, encoded_rule_evidence, oldworld, action, new_world, encoded_rules, encoded_negrules,
        predicted_world, pre_action_agent=old_agent, ground_truth_post_action_agent=new_agent,
        predicted_agent=predicted_agent,
        unobserved_code='.',
        object_count_threshold=1
    )

    # check to see we mutated nothing
    assert backup_encoded_focus_set == encoded_focus_set
    assert backup_encoded_rule_evidence == encoded_rule_evidence
    assert backup_oldworld == oldworld
    assert backup_new_world == new_world
    assert backup_encoded_rules == encoded_rules
    assert backup_encoded_negrules == encoded_negrules
    assert backup_predicted_world == predicted_world
    assert backup_old_agent == old_agent
    assert backup_new_agent == new_agent

    print(
        "About to check the rules with the action 'left', We expect 2 rules in total as there was no movement assumptions")
    rule_count = 0
    for rule in calculated_new_rules:
        if rule[0][0] == action:
            line = nace.prettyprint.prettyprint_rule(
                rule_evidence=calculated_rule_evidence,
                Hypothesis_TruthValue=nace.hypothesis.Hypothesis_TruthValue,
                rule=rule
            )
            print(line)
            prediction_state_value_deltas = rule[1]
            state_value_deltas = prediction_state_value_deltas[-1]
            rule_count += 1
    assert rule_count == 1  # we expect 1 rule
    assert state_value_deltas[0] == 0 # no change in RL reward
    assert state_value_deltas[1] == 0 # no change in termination state
    assert state_value_deltas[2] == 1  # we now have a key, indicated in agent state location 1.

    # perturb the agent so that agent precondition(s) don't match current rules and call _observe() (again),
    # and check rules were updated in the correct way
    print("INFO -------- new test start ------")
    perturbed_agent = copy.deepcopy(new_agent)
    perturbed_agent.set_values_inc_prefix((1, 0, 1))  # score += 1, still have a key (no change in keys held)

    calculated_focus_set_2, calculated_rule_evidence_2, calculated_new_rules_2, calculated_new_negrules_2 = _observe_v2(
        encoded_focus_set, encoded_rule_evidence, oldworld, action, new_world, encoded_rules, encoded_negrules,
        predicted_world, pre_action_agent=old_agent, ground_truth_post_action_agent=perturbed_agent,
        predicted_agent=predicted_agent,
        unobserved_code='.',
        object_count_threshold=1
    )

    # assert score and state0 of agent rules were both 0 in original rules.
    for rule in calculated_new_rules:
        assert rule[1][3][0] == 0 # score
        assert rule[1][3][1] == 0 # terminated
        assert rule[1][3][2] == 1 # key

    # assert score and state0 of agent rules are now both 1 in both rules (i.e. we now get reward we we get the key).
    for rule in calculated_new_rules_2:
        assert rule[1][3][0] == 1 # score
        assert rule[1][3][1] == 0 # terminated
        assert rule[1][3][2] == 1 # key
    print("test passed")



def t6_pick_up_key_increase_keys_held_is_learned():
    world_str_list = ['.............',
                      '.adddddaa....',
                      '.aeeecdbb....',
                      'AaeeeRdbb....',
                      'Aabbbbabb....',
                      'Aabbbbkbb....',
                      'Aaaaaaaaa....',
                      'AAAAAAAAA....']
    time_str_list = []
    focus_set = {'b': 53, 'd': 41, 'c': 1, 'a': 32, 'Q': 4, 'e': 57, 'O': 7, 'R': 5, 'f': 2, 'g': 1, 'h': 1, 'j': 1,
                 'N': 2, 'k': 2, 'l': 1, 'm': 2}
    active_rules = []
    rule_evidence = {}
    world_state_values = {'int_2from_rgb_list': [],
                          'int_2from_char_list': [(-1, 'A'), (-2, '.'), (3215016267, 'B'), (823474807, 'C'),
                                                  (2687125822, 'D'), (362910338, 'E'), (1758885969, 'F'),
                                                  (748094602, 'G'), (3376261930, 'H'), (3305816727, 'I'),
                                                  (854999169, 'J'), (1334950674, 'K'), (69488790, 'L'),
                                                  (2350475356, 'M'), (2829951097, 'N'), (3539787669, 'O'),
                                                  (1867891296, 'P'), (1704944625, 'Q'), (3665670421, 'R'),
                                                  (1484579719, 'a'), (1152208772, 'b'), (2624420647, 'c'),
                                                  (2204181460, 'd'), (2407174942, 'e'), (951216077, 'f'),
                                                  (4000794601, 'g'), (1483862051, 'h'), (2183995252, 'i'),
                                                  (808930194, 'j'), (1799812895, 'k'), (1406248814, 'l'),
                                                  (2624386038, 'm')], 'raw_cell_shape': (32, 32, 3),
                          'raw_nptype': "<class 'numpy.uint8'>"}
    agent = nace.agent_module.Agent(rc_loc=(3, 5), score=0.965625, terminated=0,
                                    values_excluding_prefix=[0, 0, 0, 0, 0, 0])

    # hand coded after this line
    """
    (-1, 'A'),    EXTERNAL WALL
    (-2, '.')     UNOBSERVED
    (3215016267, 'B'), facing down - green RHS
    (823474807,  'C'), facing right - green background
    (2687125822, 'D'), facing up - green tip
    (362910338,  'E'), facing right - on lava/water
    (1758885969, 'F'),  up - red tip
    (748094602,  'G'),  up - lava
    (3376261930, 'H'),  up - grey tip
    (3305816727, 'I'),  right - green door edge on RHS
    (854999169,  'J'), up
    (1334950674, 'K'),  up, grey tip
    (69488790,   'L'),  right on water/lava
    (2350475356, 'M'),  left on lava water
    (2829951097, 'N'), left 
    (3539787669, 'O'), right
    (1867891296, 'P'), down on green background
    (1704944625, 'Q'), down 
    (3665670421, 'R'), up
    (1484579719, 'a'), grey square - can be moved through
    (1152208772, 'b'), black square - can be moved through
    (2624420647, 'c'), grey key
    (2204181460, 'd'), grey square
    (2407174942, 'e'), black square
    (951216077,  'f'), grey door
    (4000794601, 'g'), grey key, dark background
    (1483862051, 'h'), grey door, darker background
    (2183995252, 'i'), dark square, light grey RHS
    (808930194, 'j'),  green key
    (1799812895, 'k'). green door

    """
    agent_char = world_str_list[agent.get_rc_loc()[0]][agent.get_rc_loc()[1]]
    world, agent_indication_raw_value_list = nace.world_module_numpy.NPWorld.from_string(
        world_str_list,
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=[agent_char], # must match actual agent char
        observed_times_str_list=[],
        int_2from_char_mapping=world_state_values["int_2from_char_list"]
    )
    world.multiworld_print([{"World": world, "Color": nace.color_codes.color_code_white_on_blue}])

    # convert focus_set and rules to use embeddings rather than char (which is used for convenience)
    new_focus_set = convert_focus_set_to_internal_mapping(focus_set, world)
    new_rules = convert_rules_to_internal_mapping(active_rules, world)
    new_rule_evidence = convert_rule_evidence_to_internal_mapping(rule_evidence, world)


    # now perform meat of the test
    # change world state to grey key (which the rules should know) check predicting picking up the key, results in a held key

    # test that the current rules suggest key count gow up if we pick up a key
    plan_actions = ['pickup']
    planworld = copy.deepcopy(world)
    planagent = copy.deepcopy(agent)
    assert sum(planagent.get_values_exc_prefix()) == 0
    for i in range(0, len(plan_actions)):
        planworld, planagent, _, __, ___, ____ = nace.nace_v3._predict_next_world_state(
            new_focus_set,
            copy.deepcopy(planworld), plan_actions[i], new_rules, planagent
        )

    # repeatedly present the same initial world and agent, and new world and agent, evidence gets more strong, but spurious rules not un-learned.
    world_str_list_with_no_agent = ['.............',
                      '.adddddaa....',
                      '.aeeeedbb....',
                      'Aaeeeedbb....',
                      'Aabbbbabb....',
                      'Aabbbbkbb....',
                      'Aaaaaaaaa....',
                      'AAAAAAAAA....']


    for i in range(15):
        desired_agent_rc_loc = (random.randint(2, 4),random.randint(2, 6))
        desired_object_rc_loc = (desired_agent_rc_loc[0]-1,desired_agent_rc_loc[1])
        below_key = random.random() > 0.5 or i <= 2
        if below_key:
            object_char = 'c' # key
            ground_truth_post_action_agent = Agent(desired_agent_rc_loc, agent.get_score(), agent.get_terminated(),
                                                   [0, 0, 0, 0, 1, 0, 0])
        else:
            object_char = 'c'  # key
            # shift the agent 1 more space down (away from key)
            desired_agent_rc_loc = (desired_agent_rc_loc[0]+1,desired_agent_rc_loc[1]) # agent is too far away for pickup to have effect
            ground_truth_post_action_agent = Agent(desired_agent_rc_loc, agent.get_score(), agent.get_terminated(),
                                                   [0, 0, 0, 0, 0, 0, 0])

        agent_char = 'R' # facing up
        input_pre_action_world, input_agent =  place_agent_and_object(world_str_list_with_no_agent, agent,  desired_agent_rc_loc, desired_object_rc_loc, object_char, agent_char, world_state_values["int_2from_char_list"] )
        post_action_ground_truth_world, _  = place_agent_and_object(world_str_list_with_no_agent, agent,  desired_agent_rc_loc, desired_object_rc_loc, 'b', agent_char, world_state_values["int_2from_char_list"] )

        world.multiworld_print([{"World": input_pre_action_world, "Color": nace.color_codes.color_code_white_on_blue, "Caption":"Pre"},
                                {"World": post_action_ground_truth_world, "Color": nace.color_codes.color_code_white_on_blue, "Caption":"Post"}])

        (
            r_used_rules,
            r_focus_set,
            r_rule_evidence,
            r_simulated_world_post_action,
            r_newrules,
            r_newnegrules,
            r_values,
            r_lastplanworld,
            r_predicted_world,
            r_stayed_the_same,
            r_stats
        ) = nace.nace_v3.nacev3_predict_and_observe(  # called _predict_and_observe in v1
                1, #time_counter,
                new_focus_set,
                new_rule_evidence,
                input_agent.get_rc_loc(),  # pre action agent location note (0,0) is top left
                input_pre_action_world, # pre_action_world,  # pre action internal world model (partially observed) before last_action is applied
                set(new_rules), #rulesin,  # used in part1 and part2
               {}, #negrules,
                'pickup', #'last_action,
            set([]), #rulesExcluded,
                post_action_ground_truth_world,  # i.e. this is the post action world model
                input_agent, #pre_action_agent,  # pre action agent
                ground_truth_post_action_agent,  # post action agent
                -1, #unobserved_code,  # value which indicates we can not see the true value of the cell
                agent_indication_raw_value_list,  # raw value of an agent, i.e. 'x', or the rgb value
                object_count_threshold=1,  # how unique values must be before they are considered. was 1, i.e. unique
                print_out_world_and_plan=True)


        # print the newly updated rules
        print("Learning cycle", i)
        for r in r_rule_evidence.keys():
            _,str_rule = convert_rules_to_char_or_int_mappings([r], world)
            if r not in new_rule_evidence:
                print("INFO.   new rule: ", str_rule ,  r_rule_evidence[r])
            else:
                if r_rule_evidence[r] != new_rule_evidence[r]:
                    print("INFO modified ", str_rule, "was", new_rule_evidence[r], "now", r_rule_evidence[r])

        for r in r_used_rules:
            if r[0][0] == 'pickup':
                _,str_rule = convert_rules_to_char_or_int_mappings([r], world)
                print("INFO.   used rule: ", str_rule  )
        for r in r_newrules:
            if r[0][0] == 'pickup':
                _,str_rule = convert_rules_to_char_or_int_mappings([r], world)
                print("INFO.   r_newrules: ", str_rule )
        for r in r_newnegrules:
            if r[0][0] == 'pickup':
                _,str_rule = convert_rules_to_char_or_int_mappings([r], world)
                print("INFO.   r_newnegrules: ", str_rule )

        # copy output over input for next learning cycle.
        new_rule_evidence = copy.deepcopy(r_rule_evidence)
        new_focus_set = copy.deepcopy(r_focus_set)
        new_rules = copy.deepcopy(list(new_rule_evidence.keys()))


    for k in new_rule_evidence.keys():
        if k[0][0] == 'pickup':
            print("DEBUG ", k)

    # assert we have a rule that when the agent is below the key, and performs 'pickup', the grey key score is incremented.
    # 2624420647 grey key, 1152208772 square that can be moved through.
    assert (('pickup', (0, 0, 0, 0, 0, 0, 0), (0, 0, 2624420647), (1, 0, 3665670421)), (0, 0, 1152208772, (0.0, 0, 0, 0, 0, 0, 1, 0))) in new_rule_evidence
    print("Test Passed")


if __name__ == "__main__":
    t6_pick_up_key_increase_keys_held_is_learned()


    t5_learn_agent_status_change_key_pickup()


    t1_simple_move_left() # passes
    t3_cup_on_table() # passes
    t2_simple_move_left_get_moved_left_and_down() # passes
    t4_learn_to_drop_agent_preconditions()




"""

Used to hold a master list of current valid actions, and convert to/from gym actions

"""
import gymnasium
import numpy as np
from gymnasium.spaces import Space

import nace.prettyprint


class NaceGymActionMapper():
    """
    Maintains mappings of actions between the 2 representations,
    - Gymnasium Actions (int.64) that may not be in sequence
        and
    - 'Nice' names that can be of any type, but being human-readable is useful.

    """

    def __init__(self,
                 gym_action_space: Space,
                 extend_beyond_configured_actions: bool,
                 gym_to_nace_name_mapping):
        """

        @param gym_action_space:
        @param extend_beyond_configured_actions: if True, if the gym environment says there are 20 actions, we will use all.
        """
        self.nace2gym = {}
        self.gym2nace = {}

        if isinstance(gym_action_space, gymnasium.spaces.box.Box):
            # Box (-2, 2, (7,), float32) - meaning 7 actions each with a value between -2 and 2 of type float 32
            pass # TBC
        elif isinstance(gym_action_space, Space):
            for gym_value, nace_value in gym_to_nace_name_mapping.items():
                self.nace2gym[nace_value] = gym_value
                self.gym2nace[gym_value] = nace_value

            if extend_beyond_configured_actions:
                # part of the action space is hard coded, the rest have to be learnt.
                for i in range(gym_action_space.n.item()):
                    gym_value = np.int64(i)
                    if gym_value not in self.gym2nace:
                        nace_action = int(i)
                        self.gym2nace[gym_value] = nace_action
                        self.nace2gym[nace_action] = gym_value

    def convertToGymAction(self, nace_action):
        nace_action_name = nace.prettyprint.get_pretty_action(nace_action)
        return np.int64(self.nace2gym[nace_action]), nace_action_name

    def get_full_action_list(self):
        # create a copy, and use a type that can be pickled.
        result = []
        result.extend(self.nace2gym.keys())
        return result



import copy


class Agent():
    """
    Holds the state of the agent, including location, and 'internal' arbitrary values (num keys held)
    """

    INDEX_SCORE = 0
    INDEX_TERMINATED = 1
    PREFIX_LENGTH = max(INDEX_SCORE, INDEX_TERMINATED) + 1

    def __init__(self, rc_loc, score, terminated=0, values_excluding_prefix=()):
        """

        @param rc_loc: row column location of the agent. Ideally should not be used in learning.
        @param score: Cumulative reward for the current episode. Can/is reset at termination
        @param terminated: True if the episode was terminated i.e. dead or reset
        @param values_excluding_prefix: Env specific values, i.e. keys held
        """
        # location is stored, but not part of agent state for purpose of learning, as ground truth is from the world.
        self.rc_loc = rc_loc
        # ideally the terminated flag would be optimised out, the agent would learn the pattern resets, but it
        # is needed to learn to avoid a reset where one is not needed.
        prefix = [score] + [int(terminated)]
        self.values = prefix + list(values_excluding_prefix)

    def __eq__(self, other):
        if not isinstance(other, Agent):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return (
                self.rc_loc == other.rc_loc and
                self.values == other.values
        )

    @staticmethod
    def new_default_instance(rc_loc):
        return Agent(rc_loc, 0, 0, [])

    @staticmethod
    def get_prefix_length():
        return Agent.PREFIX_LENGTH

    def set_rc_loc(self, rc_loc):
        self.rc_loc = rc_loc

    def get_rc_loc(self):
        return copy.deepcopy(self.rc_loc)

    def set_score(self, score):
        self.values[Agent.INDEX_SCORE] = score

    def get_score(self):
        return self.values[Agent.INDEX_SCORE]

    def get_terminated(self):
        return self.values[Agent.INDEX_TERMINATED]

    def get_values_for_precondition(self):
        # Return all values EXCLUDING score, which matches a rule's PRE-CONDITION,
        # but note the effect includes the score.

        # Contains all the values that are needed to be included to see if agent state is the same for the
        # purposes of planning, i.e. includes terminated, but not score, as being in the same place with the same
        # things but with a greater score should be the same. This is used in the rule preconditions.

        # NOTE: adding anything into the agents values that has a large number of
        # discrete values explodes the size of the stored rules.
        return copy.deepcopy(self.values[1:])

    def get_values_exc_prefix(self):
        # this is used to generate code for the purposes of unit tests, and matches a param in the constructor.

        # NOTE: adding anything into the agents values that has a large number of
        # discrete values explodes the size of the stored rules.
        return copy.deepcopy(self.values[2:])

    def get_values_inc_prefix(self):
        # NOTE: extending the length of the values may explode the size of the planning space.
        return copy.deepcopy(self.values)

    def set_values_inc_prefix(self, values):
        self.values = list(copy.deepcopy(values))

    def accumulate_values_incl_prefix(self, values):
        for i in range(len(values)):
            if i < len(self.values):
                self.values[i] += values[i]



import numpy as np

from world_module_numpy import NPWorld


def run_1():
    # test a known grid world renders correctly and the robot is in the correct place.
    gym_world = 'o  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\nx  C  C  C  C  C  C  C  C  C  C  T\n\n'
    world = NPWorld(with_observed_time=True, name="")
    world.update_world_from_ground_truth_NPArray(gym_world=gym_world)
    world_as_string = world.board_as_string()
    assert world_as_string[61] == 'x'
    print(world_as_string)


def run_2_differences():
    gym_world1 = 'o  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  x  o  o  o  o  o\no  C  C  C  C  C  C  C  C  C  C  T\n\n'
    world1 = NPWorld(with_observed_time=True, name="")
    world1.update_world_from_ground_truth_NPArray(gym_world=gym_world1)

    gym_world2 = 'o  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\nx  C  C  C  C  C  C  C  C  C  C  T\n\n'
    world2 = NPWorld(with_observed_time=True, name="")
    world2.update_world_from_ground_truth_NPArray(gym_world=gym_world2)

    row_indexes, column_indexes = world1.get_list_of_differences(world2)

    assert world1.get_difference_count(world2) == len(column_indexes)

    print("World 1")
    print(world1.board_as_string())
    print("World 2")
    print(world2.board_as_string())

    print("differences. rows:", row_indexes, "columns:", column_indexes)
    assert world1.board[row_indexes[0], column_indexes[0]] != world2.board[row_indexes[0], column_indexes[0]]
    assert world1.board[row_indexes[1], column_indexes[1]] != world2.board[row_indexes[1], column_indexes[1]]


def run_3_hashcodes():
    gym_world1 = 'o  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  x  o  o  o  o  o\no  C  C  C  C  C  C  C  C  C  C  T\n\n'
    gym_world2 = 'o  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\nx  C  C  C  C  C  C  C  C  C  C  T\n\n'
    #                                                                                                       ^diff             ^diff

    for with_observed_time in [True, False]:
        for (view_dist_x, view_dist_y) in [(3, 2), (10, 10), (100, 100)]:
            world1 = NPWorld(with_observed_time=with_observed_time, name="", view_dist_x=view_dist_x,
                             view_dist_y=view_dist_y)
            world1.update_world_from_ground_truth_NPArray(gym_world=gym_world1)

            world2 = NPWorld(with_observed_time=with_observed_time, name="", view_dist_x=view_dist_x,
                             view_dist_y=view_dist_y)
            world2.update_world_from_ground_truth_NPArray(gym_world=gym_world2)

            world1_duplicate = NPWorld(with_observed_time=with_observed_time, name="", view_dist_x=view_dist_x,
                                       view_dist_y=view_dist_y)
            world1_duplicate.update_world_from_ground_truth_NPArray(gym_world=gym_world1)

            assert world1.get_board_hashcode() != world2.get_board_hashcode()
            assert world1.get_board_hashcode() == world1_duplicate.get_board_hashcode()

            world1_duplicate.set_embedded_val_rc(2, 2, 'p')
            assert world1.get_board_hashcode() != world1_duplicate.get_board_hashcode()


def run_4_to_from_string():
    gym_world1 = 'o  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  o  o  o  o  o  o\no  o  o  o  o  o  x  o  o  o  o  o\no  C  C  C  C  C  C  C  C  C  C  T\n\n'
    world1 = NPWorld(with_observed_time=True, name="unittest")
    world1.update_world_from_ground_truth_NPArray(gym_world=gym_world1)
    intermediate_fmt = world1.to_string_list()
    rebuilt = NPWorld.from_string_list(intermediate_fmt)
    assert np.array_equal(world1.times, rebuilt.times)
    assert np.array_equal(world1.board, rebuilt.board)




if __name__ == "__main__":
    run_1()
    run_2_differences()
    run_3_hashcodes()
    run_4_to_from_string()
    run_5_convert_gym_world()


import copy
import json
import os.path

import numpy as np
import xxhash
# from nace.numpy_utils import convert_string_list_to_cells
from bidict import bidict

import nace
import nace.color_codes
import nace.numpy_utils

try:
    import cv2  # pip install opencv-python

    is_cv2_installed = True
except ImportError as e:
    is_cv2_installed = False

EMBEDDED_WALL_CODE = -1
UNOBSERVED_BOARD_VALUE = -2


class NPWorld():

    def __init__(self, with_observed_time, name, width=None, height=None,
                 initial_value=UNOBSERVED_BOARD_VALUE,
                 view_dist_x=3, view_dist_y=2,
                 store_raw_cell_data=True,
                 raw_cell_data_location='./data/tmp',
                 rgb_to_char_mapping={},  # used to force a raw value to a certain char for ease of debugging
                 raw_nptype=None,
                 raw_cell_shape=None,
                 agent_indication_raw_value_list=[],  # if known, what the agent can look like is very handy
                 user_set_int_2from_char_mapping=[],  #
                 ):
        self.name = name
        if width is None or height is None:
            self.board = np.zeros((0, 0), dtype=np.int64)
        else:
            self.board = np.zeros((width, height), dtype=np.int64)
            self.board[:] = initial_value

        self.initial_value = initial_value
        self.with_observed_time = with_observed_time

        self.times = np.zeros((0, 0), dtype=np.float16)  # this seems expensive in space, a float16 for each cell. hmmm
        self.times[:] = float('-inf')
        # self.agent_location_column = -1 # if cells, in cell units
        # self.agent_location_row = -1 # if cells, in cell units
        self.view_dist_x = view_dist_x
        self.view_dist_y = view_dist_y
        self.raw_cell_data_location = raw_cell_data_location
        self.store_raw_cell_data = store_raw_cell_data
        self.debug_board = ""

        # Type of the cell in raw input, often set latter, set once, can not be changed.
        self.raw_nptype = raw_nptype  # Must be set if rgb_to_char_mapping is passed in.
        self.raw_cell_shape = raw_cell_shape  # Must be set if rgb_to_char_mapping is passed in.
        if rgb_to_char_mapping is not None and rgb_to_char_mapping != {} and (
                raw_nptype == None or raw_cell_shape is None):
            raise ("If rgb_to_char_mapping is passed in, raw_nptype AND raw_cell_shape must be set as well.")

        self.int_2from_rgb_mapping = bidict()  # updated whenever a hash code is calculated
        self.int_2from_char_mapping = bidict()  # updated when the char encoding is needed.

        # Set special values
        self.int_2from_char_mapping[EMBEDDED_WALL_CODE] = 'A'
        self.int_2from_char_mapping[UNOBSERVED_BOARD_VALUE] = '.'

        for (k, v) in user_set_int_2from_char_mapping:
            assert isinstance(k, int)
            assert isinstance(v, str)
            self.int_2from_char_mapping[k] = v

        # convert external agent images to internal values
        agent_indication_raw_value_list = self._convert_raw_to_embedded_values(agent_indication_raw_value_list)

        # add any hard coded values the user has passes in (i.e. when testing and setting the map)
        for rgb_bytes, ch in rgb_to_char_mapping.items():
            if ch in self.int_2from_char_mapping.values():
                val = self.int_2from_char_mapping.inverse[ch]
            else:
                rgb_cell = self._convert_bytes_to_rgb_ndarray(rgb_bytes)
                val = self.get_val_for_rgb_cell(rgb_cell)

            self.int_2from_rgb_mapping[val] = rgb_bytes
            self.int_2from_char_mapping[val] = ch
            assert ch == self._decode_val_to_char(val, agent_indication_raw_value_list)

    def __eq__(self, other):
        if not isinstance(other, NPWorld):
            # don't attempt to compare against unrelated types
            return NotImplemented

        comparison = self.board == other.board
        board_matches = comparison.all()
        comparison = self.times == other.times
        times_match = comparison.all()

        return (
                board_matches and
                self.initial_value == other.initial_value and
                self.with_observed_time == other.with_observed_time and
                times_match and
                self.view_dist_x == other.view_dist_x and
                self.view_dist_y == other.view_dist_y and
                self.raw_cell_data_location == other.raw_cell_data_location and
                self.store_raw_cell_data == other.store_raw_cell_data and
                self.debug_board == other.debug_board and
                self.raw_nptype == other.raw_nptype and
                self.raw_cell_shape == other.raw_cell_shape and
                self.int_2from_rgb_mapping == other.int_2from_rgb_mapping and
                self.int_2from_char_mapping == other.int_2from_char_mapping
        )

    # ======= START of Value / rgb array / Char Mapping Calculations ========== #

    """
    must have embedded wall value, so it can be added round board
    must have unobserved value, so it can be treated differently (should prevent bugs)
    
    if we want these to be assigned to random rgb values for testing, then we need RGB <> Int <> Char mappings
    
    """

    def _get_new_non_reserved_char(self, v):
        # calculate a new non reserved char (lowercase and beyond)
        max_char = chr(ord('a') - 1)  # one before 'a' so first char is 'a'
        if len(self.int_2from_char_mapping.values()) > 0:
            max_char = max(max_char, max(self.int_2from_char_mapping.values()))
        new_char = chr(ord(max_char) + 1)
        assert v not in self.int_2from_char_mapping
        if isinstance(v, np.int64):
            self.int_2from_char_mapping[v.item()] = new_char
        else:
            assert isinstance(v, int)
            self.int_2from_char_mapping[v] = new_char

        return new_char

    def _decode_val_to_char(self, v, agent_indication_embedded_value_list):
        # negative values, and uppercase char are reserved.
        # int_2from_char_mapping and int_2from_rgb are first class citizens

        # if v == EMBEDDED_WALL_CODE:
        #     return 'A'
        # if v == UNOBSERVED_BOARD_VALUE:
        #     return '.'

        if v in self.int_2from_char_mapping:
            return self.int_2from_char_mapping[v]

        # ensure the embedded agent values have Char encodings
        for embedded_value in agent_indication_embedded_value_list:
            if embedded_value not in self.int_2from_char_mapping:  # can't be due to above logic
                max_char = 'A'  # A reserved for walls, B-Z reserved for agents
                if len(self.int_2from_char_mapping.values()) > 0:
                    max_char = max(max_char,
                                   max([ch for ch in self.int_2from_char_mapping.inverse.keys() if ch <= 'Z']))
                if max_char == 'Z':
                    # run out of B-Z for agent, must now use general range
                    pass
                else:
                    new_char = chr(ord(max_char) + 1)
                    self.int_2from_char_mapping[embedded_value] = new_char

        # value may have been added in the last block, check again and return if present
        if v in self.int_2from_char_mapping:
            return self.int_2from_char_mapping[v]

        # calculate a new non reserved char (lowercase and beyond)
        new_char = self._get_new_non_reserved_char(v)

        return new_char

    def _convert_bytes_to_rgb_ndarray(self, rgb_bytes):

        assert self.raw_nptype is not None
        deserialized_bytes = np.frombuffer(rgb_bytes, dtype=self.raw_nptype)
        newshape = (self.raw_cell_shape[0], self.raw_cell_shape[1], self.raw_cell_shape[2])
        rgb = np.reshape(deserialized_bytes, newshape=newshape)
        return rgb

    def get_val_for_char(self, c):
        if c in self.int_2from_char_mapping.inverse:
            return self.int_2from_char_mapping.inverse[c]
        raise Exception("Mapping not found for chr '" + str(c) + "'")

    def get_val_to_char_mappings(self, list_of_extra_values=[]):
        # this routine could be simplified, too much looping.
        # return the mappings as of now (they can be updated of course)
        char_set = set([ch for ch in self.debug_board])
        result = {}
        for ch in char_set:
            if ch != '\n':
                val = self.get_val_for_char(ch)
                result[val] = ch
        for val in list_of_extra_values:
            if val in self.int_2from_char_mapping:
                ch = self.int_2from_char_mapping[val]
                result[val] = ch
            else:
                # this is probably caused by not knowing whether the code (and image) in question is an agent or
                # not, and we do not know how to encode it.
                # calculate a new non reserved char (lowercase and beyond)
                new_char = self._get_new_non_reserved_char(val)
                print("WARN can not convert "+str(val)+" to char for debugging purposes. Created a new non reserved char. ", new_char)
                result[val] = new_char

        for val in self.int_2from_char_mapping:
            result[val] = self.int_2from_char_mapping[val]

        return result

    def get_val_for_rgb_cell(self, rgb_cell: np.ndarray):
        # generates an embedded int value for a rgb cell and stores it in self.int_2from_rgb_mapping
        # for fast lookup later

        if self.raw_cell_shape is None:
            self.raw_cell_shape = rgb_cell.shape
        if self.raw_nptype is None:
            if len(rgb_cell.shape) == 3:
                self.raw_nptype = type(rgb_cell[0][0][0])
            elif len(rgb_cell.shape) == 2:
                self.raw_nptype = type(rgb_cell[0][0])


        for dim in range(len(rgb_cell.shape)):
            if rgb_cell.shape[dim] != self.raw_cell_shape[dim]:
                print("DEBUG rgb_cell.shape", rgb_cell.shape)
                print("DEBUG self.raw_cell_shape", self.raw_cell_shape)
            assert rgb_cell.shape[dim] == self.raw_cell_shape[dim]

        if len(rgb_cell.shape) == 3:
            assert isinstance(rgb_cell[0][0][0], self.raw_nptype)
        elif len(rgb_cell.shape) == 2:
            assert isinstance(rgb_cell[0][0], self.raw_nptype)

        rgb_bytes = rgb_cell.data.tobytes()
        if rgb_bytes in self.int_2from_rgb_mapping.values():  # return previous values (alows us to overide hashcode for special values
            return self.int_2from_rgb_mapping.inverse[rgb_bytes]

        hc = xxhash.xxh32(rgb_bytes).intdigest()  # should be repeatable between restarts.
        # this is to fix the issue that the hash code are not deterministic, between restarts
        # they can change, this means we can not serialize a knowledge base of rules.
        hci = int(hc)
        # hci = hci % 2147483647 # np int limit. wrap values (more chance of collision, but can be stored in int)

        if self.raw_cell_shape is None:
            self.raw_cell_shape = (rgb_cell.shape[0], rgb_cell.shape[1], rgb_cell.shape[2])
            if len(self.raw_cell_shape) >= 3:
                assert self.raw_cell_shape[2] == 3
        else:
            for dim in range(len(rgb_cell.shape)):
                assert (self.raw_cell_shape[dim] == rgb_cell.shape[dim])

        return hci

    # ======= END of Value / rgb array / Char Mapping Calculations ========== #

    def get_unobserved_code(self):
        return UNOBSERVED_BOARD_VALUE

    def get_char_at_rc(self, row, col, agent_indication_embedded_value_list):
        " We store integers in each cell, this decodes to 'nice' char values"
        return self._decode_val_to_char(self.board[row, col], agent_indication_embedded_value_list)

    def set_embedded_val_rc(self, row: int, col: int, val: int):
        self.board[row, col] = val

    def get_embedded_val_rc(self, row: int, col: int):
        return self.board[row, col].item()

    def get_time_at_rc(self, row, col):
        return self.times[row, col]

    def get_newest_time(self):
        """
        @return: the must recent, up-to-date time stored in the world.
        """
        if self.with_observed_time:
            if self.times.shape[0] > 0:
                return max(0, self.times.max())
        return 0

    def get_locations_updated_tminus(self, time_delta=1):
        newest_time = self.get_newest_time()
        search_time = newest_time - time_delta
        locations = [(r.item(), c.item()) for (r, c) in zip(*np.where(self.times == search_time))]
        return locations

    def get_board_line(self, row_number, color, agent_indication_embedded_value_list, no_color_on_oldest=False):
        if self.with_observed_time and self.times.size > 0:
            oldest_time = self.times.min().item()
        else:
            oldest_time = float('-inf')
        line = ""
        color_count = 0
        if row_number < self.board.shape[0]:
            for col in range(self.board.shape[1]):
                if self.with_observed_time and self.times[row_number][col] == oldest_time and no_color_on_oldest:
                    line += nace.color_codes.color_code_black_on_white
                    color_count += 1
                line += self.get_char_at_rc(row_number, col, agent_indication_embedded_value_list)

                if self.with_observed_time and self.times[row_number][col] == oldest_time:
                    if color != None:
                        line += color
        return line, self.board.shape[1]

    def get_in_nace_board_format(self):
        board = []
        for row in range(self.board.shape[0]):
            line = []
            for col in range(self.board.shape[1]):
                line.append(self.get_char_at_rc(row, col))
            board.append(line)
        return board

    def _reset_non_board(self):
        # set the times and other values to the default
        if self.with_observed_time:
            self.times = np.zeros(self.board.shape, dtype=np.float16)  # this seems expensive in space. hmmm
            self.times[:] = float('-inf')

    def _set_board_size(self, height, width):
        prior_board_shape = self.board.shape
        size_changed = False
        if prior_board_shape == (0, 0):
            # we need to reset all other values
            self.board = np.zeros((height, width), dtype=np.int64)
            self.board[:] = self.initial_value
            self._reset_non_board()
            size_changed = True

        return prior_board_shape, size_changed

    def _get_windowed_rc_locations(self, rc_locations, whole_screen_raw_width, whole_screen_raw_height, cell_width=1,
                                   cell_height=1):
        """
        Return a list of points within view distance of the agents(s) rc_locations
        """
        if rc_locations == []:
            return []  # speed optimisation
        rc_points_to_copy = []
        rows = list(range(self.view_dist_y * 2 + 1))
        columns = list(range(self.view_dist_x * 2 + 1))

        for r_delta in rows:
            for c_delta in columns:
                for rc_loc in rc_locations:
                    if len(rc_loc) < 2:
                        raise Exception("Type logic Error")
                    if isinstance(rc_loc[1], tuple) or isinstance(r_delta, tuple) or isinstance(self.view_dist_y,
                                                                                                tuple):
                        print("breakpoint")
                    r = rc_loc[0] + r_delta - self.view_dist_y
                    c = rc_loc[1] + c_delta - self.view_dist_x
                    if r >= 0 and r * cell_height < whole_screen_raw_height and \
                            c >= 0 and c * cell_width < whole_screen_raw_width:
                        rc_points_to_copy.append((r, c))
        rc_points_to_copy = list(set(rc_points_to_copy))  # ensure unique
        return rc_points_to_copy

    @staticmethod
    def extract_cell(r: int, c: int, observed_ndarray: np.ndarray, cell_shape_rc=(1, 1)):
        assert (r + 1) * cell_shape_rc[0] <= observed_ndarray.shape[0] + 1
        assert (c + 1) * cell_shape_rc[1] <= observed_ndarray.shape[1] + 1
        cell = observed_ndarray[r * cell_shape_rc[0]:(r + 1) * cell_shape_rc[0],
               c * cell_shape_rc[1]:(c + 1) * cell_shape_rc[1]
               ]
        return cell

    def _convert_raw_to_embedded_values(self, agent_indication_raw_value_list):
        agent_indication_embedded_value_list = [self.get_val_for_rgb_cell(cell) for cell in
                                                agent_indication_raw_value_list]
        return agent_indication_embedded_value_list

    def _copy_mappings_from_source_to_self(self, v: int, source_world):
        if self.raw_cell_shape is not None and source_world.raw_cell_shape is not None:
            if self.raw_cell_shape != source_world.raw_cell_shape:
                raise Exception("Cell shape already set, and can not be changed")

        if self.raw_cell_shape is None and source_world.raw_cell_shape is not None:
            self.raw_cell_shape = source_world.raw_cell_shape

        if v not in self.int_2from_rgb_mapping:
            if v in source_world.int_2from_rgb_mapping:
                rgb_bytes = source_world.int_2from_rgb_mapping[v]

                self.int_2from_rgb_mapping[v.item()] = rgb_bytes
                self.int_2from_char_mapping[v.item()] = source_world.int_2from_char_mapping[v]

    def _copy_from_source_world_to_self(self, rc_points_to_copy: list, source_world, time_counter: int):
        modified_count = 0
        for (r, c) in rc_points_to_copy:
            if source_world.board[r][c] != UNOBSERVED_BOARD_VALUE:
                v = source_world.board[r][c]
                if self.board[r][c] != v:
                    self.board[r][c] = v
                    self._copy_mappings_from_source_to_self(v, source_world)
                    modified_count += 1
                if self.with_observed_time:
                    self.times[r][c] = time_counter
        return modified_count

    def update_world_from_ground_truth(
            self,
            time_counter: int,
            external_ground_truth_world_model,  # type:NPWorld
            rc_locations,  # list of locations, around which data is copied from external to internal world.
            # (Usually just where the agent is)
            agent_indication_raw_value_list: list):

        rc_locations = copy.deepcopy(rc_locations)
        height = external_ground_truth_world_model.board.shape[0]
        width = external_ground_truth_world_model.board.shape[1]

        previous_size, size_changed = self._set_board_size(
            height=height,
            width=width)
        if size_changed:
            if previous_size != (0, 0):
                print("ERROR: should not change in size. ")
                raise Exception("Changing board size once initialised not supported.")

        pre_action_internal_world = copy.deepcopy(self)

        # add the last known place of the agents to the list of update locations, as the agent may have 'jumped'
        agent_last_rc_location = self.extract_agent_location_raw(agent_indication_raw_value_list)
        if len(agent_last_rc_location) > 0:
            assert isinstance(agent_last_rc_location, tuple)
            if agent_last_rc_location != (-1, -1):
                rc_locations.append(agent_last_rc_location)
            else:
                print("ERROR no know last location. Maybe there is no agent. Check last screen shot. And that all images of agent are in agents dir (along with matching .npy file). rc_locations=",str(rc_locations))

        if len(agent_last_rc_location) == 0:
            print("ERROR - logic error? will the next line now fail?")

        # Filter out negative (illegal) values (which indicate location is unknown)
        rc_locations = [t for t in rc_locations if t is not None]
        rc_locations = [(r, c) for (r, c) in rc_locations if r >= 0 and c >= 0]

        # uses view distance, and only returns points in that distance
        rc_points_to_copy = self._get_windowed_rc_locations(rc_locations, width, height)
        modified_count = self._copy_from_source_world_to_self(rc_points_to_copy, external_ground_truth_world_model,
                                                              time_counter)

        if modified_count == 0:
            pass

        agent_indication_embedded_value_list = self._convert_raw_to_embedded_values(agent_indication_raw_value_list)
        self.debug_board = self.board_as_string(agent_indication_embedded_value_list)

        # self.multiworld_print([{"World": self, "Caption":self.name, "Color":nace.color_codes.color_code_white_on_red}])

        return modified_count, pre_action_internal_world

    def _save_new_cells(self, observed_ndarray: np.ndarray, cell_shape_rc: tuple,
                        agent_indication_raw_value_list: list):

        if len(agent_indication_raw_value_list) > 0 and type(agent_indication_raw_value_list[0]) == str:
            # we are using char input rather than rgb, do not save to disk.
            return

        """ Save cell images to disk for ease of debugging"""
        agent_indication_embedded_value_list = list(set([
            self.get_val_for_rgb_cell(cell) for cell in agent_indication_raw_value_list if
            cell.shape == self.raw_cell_shape
        ]))
        for embedded_r in range(int(observed_ndarray.shape[0] / cell_shape_rc[0])):
            for embedded_c in range(int(observed_ndarray.shape[1] / cell_shape_rc[1])):
                # embedded_r, embedded_c are in input space units, i.e. no walls, and in cells
                cell = NPWorld.extract_cell(embedded_r, embedded_c, observed_ndarray, cell_shape_rc=cell_shape_rc)
                cell_embedded_value = self.get_val_for_rgb_cell(cell)
                if self.store_raw_cell_data:
                    if cell_embedded_value not in agent_indication_embedded_value_list:
                        # don't store the images of the known agents, so it is easy to see if new agent values exisit ( ma nually)
                        fn = os.path.join(self.raw_cell_data_location, str(cell_embedded_value) + '.png')
                        if os.path.exists(self.raw_cell_data_location):
                            if not os.path.isfile(fn):
                                if is_cv2_installed:
                                    cv2.imwrite(fn, cell)
                                else:
                                    print("WARN: not saving image as cv2 not installed.")

                            fn = os.path.join(self.raw_cell_data_location, str(cell_embedded_value) + '.npy')
                            if not os.path.isfile(fn):
                                np.save(fn, cell)  # .npy extension is added if not given
        if self.store_raw_cell_data:
            fn = os.path.join(self.raw_cell_data_location, 'screen' + '.png')
            if is_cv2_installed:
                cv2.imwrite(fn, observed_ndarray)
            else:
                print("WARN: not saving image as cv2 not installed.")

    def update_world_from_ground_truth_NPArray(
            self,
            observed_ndarray: np.ndarray,
            update_mode: str,  # 'ALL' or 'VIEW'
            cell_shape_rc: tuple,  # shape of each cell
            add_surrounding_walls: bool = True,
            wall_code: int = EMBEDDED_WALL_CODE,  # int for ansi
            agent_indication_raw_value_list=['B'],
            observed_at: float = float('-inf')):
        """

        Notes:
        - observed_world may be of a larger size than the board, i.e. each 60x60x3 cell may map to a single board
            cell/value.
        - observed_world may not have a 'wall' around it, which can be added if configured to make later logic
             simplier.

        @param observed_ndarray:
                if rgb_shape is None then ndarray shape (1,n,m)
                else (n,m,3)
        @param add_surrounding_walls: bool
        @param wall_code: call value of optional surrounding wall
        @param raw_agent_indication_value: value indication current cell contains an agent
        @param observed_at: the time of the observation
        @return:
        """

        assert update_mode in ['ALL', 'VIEW'], ("update mode must be one of ALL (updates the whole board), or 'VIEW' "
                                                "(updates cells close to agents only)")

        pre_update_world = copy.deepcopy(self)
        modified_count = 0
        modified_locations_embedded_rc = []
        if self.store_raw_cell_data: # dv added 5 feb 2024
            self._save_new_cells(observed_ndarray, cell_shape_rc, agent_indication_raw_value_list)

        if add_surrounding_walls:
            double_wall_size = 2
            wall_thickness = 1
        else:
            double_wall_size = 0
            wall_thickness = 0

        if self.raw_cell_shape is None:
            self.raw_cell_shape = cell_shape_rc
        else:
            if self.raw_cell_shape != cell_shape_rc:
                raise Exception("changing board shape not supported.")

        # a group of elements in observed maps to 1 in the internal
        prior_board_shape, size_changed = self._set_board_size(
            height=int(observed_ndarray.shape[0] / self.raw_cell_shape[0]) + double_wall_size,
            width=int(observed_ndarray.shape[1] / self.raw_cell_shape[1]) + double_wall_size)

        if prior_board_shape != (0, 0):
            if prior_board_shape != self.board.shape:
                raise Exception("Changing board size/shape in flight not yet supported.")

        if add_surrounding_walls:
            if self.board[0][0] != wall_code:
                for r in range(self.board.shape[0]):
                    self.board[r][0] = wall_code
                    self.board[r][-1] = wall_code
                for c in range(self.board.shape[1]):
                    self.board[0][c] = wall_code
                    self.board[-1][c] = wall_code

        agent_indication_embedded_value_list = [self.get_val_for_rgb_cell(cell) for cell in
                                                agent_indication_raw_value_list if cell.shape == self.raw_cell_shape]
        # line below: 3pm 10/Dec was correct. 3:33 below incorrect mixed pre and post
        agent_current_rc_locations_in_embedded_space = self.extract_agent_locations_raw(observed_ndarray,
                                                                                        agent_indication_embedded_value_list,
                                                                                        wall_thickness)

        # check the not yet updated board for the last location
        # block below 3pm 10/Dec was correct. 3:33 below incorrect mixed pre and post
        agent_previous_rc_locations_in_embedded_space = []
        for r in range(self.board.shape[0]):
            for c in range(self.board.shape[1]):
                for raw_agent_indication_value in agent_indication_raw_value_list:
                    if raw_agent_indication_value.shape == self.raw_cell_shape:
                        embedded_agent_indication_value = self.get_val_for_rgb_cell(raw_agent_indication_value)
                        if self.board[r, c] == embedded_agent_indication_value:
                            if (r, c) not in agent_previous_rc_locations_in_embedded_space:
                                agent_previous_rc_locations_in_embedded_space.append((r, c))
        #

        # testing the below block to find agent
        # 3pm 10/Dec next line was **incorrect**
        ____post_update_agent_location_list = self.extract_agent_locations_embedded(
            agent_indication_embedded_value_list)

        # find the agents location in new observed world
        # check the passed in observed array to get the current location
        # 3pm 10/Dec was correct, but matches code above
        # agent_current_rc_locations_in_embedded_space = []
        # for raw_agent_indication_value in agent_indication_raw_value_list:
        #     if raw_agent_indication_value.shape == self.raw_cell_shape:
        #         embedded_agent_indication_value = self.get_val_for_rgb_cell(raw_agent_indication_value)
        #         for r_without_wall in range(int(observed_ndarray.shape[0] / self.raw_cell_shape[0])):
        #             for c_without_wall in range(int(observed_ndarray.shape[1] / self.raw_cell_shape[1])):
        #                 cell = NPWorld.extract_cell(r_without_wall, c_without_wall, observed_ndarray, cell_shape_rc)
        #                 v = self.get_val_for_rgb_cell(cell)
        #                 if embedded_agent_indication_value == v:
        #                     if (r_without_wall + wall_thickness,
        #                         c_without_wall + wall_thickness) not in agent_current_rc_locations_in_embedded_space:
        #                         agent_current_rc_locations_in_embedded_space.append(
        #                             (r_without_wall + wall_thickness, c_without_wall + wall_thickness))

        agent_current_and_previous_rc_locations_embedded_space = []
        agent_current_and_previous_rc_locations_embedded_space.extend(agent_current_rc_locations_in_embedded_space)
        agent_current_and_previous_rc_locations_embedded_space.extend(agent_previous_rc_locations_in_embedded_space)
        if len(agent_current_and_previous_rc_locations_embedded_space) == 0:
            # we have no agent. perhaps their appearance changed. get the location(s)
            # of cells that changed.
            # THIS CODE BLOCK may not be needed.
            # NOTE: this violates the view dist, we should only check cells around the last known agent, but if we have no idea where agent is ....
            possible_agent_locations = []
            for r in range(int(observed_ndarray.shape[0] / self.raw_cell_shape[0])):
                for c in range(int(observed_ndarray.shape[1] / self.raw_cell_shape[1])):
                    observed_cell = NPWorld.extract_cell(r, c, observed_ndarray, cell_shape_rc)
                    observed_cell_embedded_value = self.get_val_for_rgb_cell(observed_cell)
                    if observed_cell_embedded_value in agent_indication_embedded_value_list:
                        possible_agent_locations.append((r + wall_thickness, c + wall_thickness))
            if 0 < len(possible_agent_locations) <= 2:  # NOTE: limitation allows at most 2 agents
                agent_current_and_previous_rc_locations_embedded_space.extend(possible_agent_locations)
        if len(agent_current_and_previous_rc_locations_embedded_space) == 0:
            print("Could not find agent, or agent changed embedded value, i.e. looks different")

        if update_mode == 'VIEW':
            # find points to copy from external view to internal view (only that around the agent)
            # embedded_space means with walls
            rc_points_to_copy_in_embedded_space = self._get_windowed_rc_locations(
                agent_current_and_previous_rc_locations_embedded_space,
                whole_screen_raw_width=observed_ndarray.shape[1] + (2 * wall_thickness * self.raw_cell_shape[1]),
                whole_screen_raw_height=observed_ndarray.shape[0] + (2 * wall_thickness * self.raw_cell_shape[0]),
                cell_width=self.raw_cell_shape[1],
                cell_height=self.raw_cell_shape[0]
            )

        if update_mode == 'ALL':
            # find points to copy from external view to internal view (all)
            rc_points_to_copy_in_embedded_space = []
            r_limit = int(observed_ndarray.shape[0] / cell_shape_rc[0])
            c_limit = int(observed_ndarray.shape[1] / cell_shape_rc[1])
            for r in range(r_limit):
                for c in range(c_limit):
                    rc_points_to_copy_in_embedded_space.append((r + wall_thickness, c + wall_thickness))

        for dim in [0, 1]:
            assert int(observed_ndarray.shape[dim] / cell_shape_rc[dim]) == self.board.shape[dim] - (2 * wall_thickness)

        for (embedded_r,
             embedded_c) in rc_points_to_copy_in_embedded_space:  # in embedded space, with walls in units of cells
            if wall_thickness == 0 or (
                    int(observed_ndarray.shape[0] / cell_shape_rc[0]) >= embedded_r >= wall_thickness and
                    int(observed_ndarray.shape[1] / cell_shape_rc[1]) >= embedded_c >= wall_thickness):
                observed_cell = NPWorld.extract_cell(embedded_r - wall_thickness, embedded_c - wall_thickness,
                                                     observed_ndarray, cell_shape_rc=cell_shape_rc)
                embedded_value_of_observed_cell = self.get_val_for_rgb_cell(observed_cell)
                if embedded_value_of_observed_cell != self.initial_value:
                    r = embedded_r
                    c = embedded_c
                    if self.board[r][c] != embedded_value_of_observed_cell:
                        self.board[r][c] = embedded_value_of_observed_cell
                        modified_count += 1
                        modified_locations_embedded_rc.append((r, c))
                    if self.with_observed_time:
                        self.times[r][c] = float(observed_at)
                else:  # unobserved
                    pass  # leave the observed time, and board value as they are.

        self.debug_board = self.board_as_string(agent_indication_embedded_value_list)
        # self.multiworld_print([{"World": self, "Caption":self.name, "Color":nace.color_codes.color_code_white_on_red}])

        return agent_previous_rc_locations_in_embedded_space, agent_current_rc_locations_in_embedded_space, modified_count, pre_update_world

    def extract_agent_location_raw(
            self,
            agent_indication_raw_value_list):  # note there may be more than 1 agent at some stage.
        """
        return in row colum format (r,c)
        This is used by agents that need to store their location in their internal state.
        """
        agent_indication_embedded_value_list = [self.get_val_for_rgb_cell(cell) for cell in
                                                agent_indication_raw_value_list if cell.shape == self.raw_cell_shape]
        agent_locations_list = self.extract_agent_locations_embedded(agent_indication_embedded_value_list)
        if len(agent_locations_list) > 0:
            return (agent_locations_list[0][0], agent_locations_list[0][1])
        else:
            return (-1, -1)

    def extract_agent_locations_raw(self, observed_ndarray, agent_indication_embedded_value_list, wall_thickness):
        agent_locations = []
        for r in range(int(observed_ndarray.shape[0] / self.raw_cell_shape[0])):
            for c in range(int(observed_ndarray.shape[1] / self.raw_cell_shape[1])):
                observed_cell = NPWorld.extract_cell(r, c, observed_ndarray, self.raw_cell_shape)
                observed_cell_embedded_value = self.get_val_for_rgb_cell(observed_cell)
                if observed_cell_embedded_value in agent_indication_embedded_value_list:
                    agent_locations.append((r + wall_thickness, c + wall_thickness))
        return agent_locations

    def extract_agent_locations_embedded(self, agent_indication_embedded_value_list):
        agent_last_rc_locations = []
        agent_indication_embedded_value_list = list(set(agent_indication_embedded_value_list))
        for agent_value in agent_indication_embedded_value_list:
            locations = [(r.item(), c.item()) for (r, c) in zip(*np.where(self.board == agent_value))]
            if len(locations) > 0 and isinstance(locations[0], int):
                print("place breakpoint here")
            agent_last_rc_locations.extend(locations)
        return agent_last_rc_locations

    def get_list_of_differences(self, world2):
        differences = self.board != world2.board
        (row_indexes, column_indexes) = np.nonzero(
            differences)  # return 2 vectors, with all the y values in one, and x in the other
        return row_indexes, column_indexes

    def get_difference_count(self, world2):
        if world2 is None:
            return 0
        differences = self.board != world2.board
        (row_indexes, column_indexes) = np.nonzero(
            differences)  # return 2 vectors, with all the y values in one, and x in the other
        return len(row_indexes)


    def get_board_hashcode(self):
        """
        @return:
        """
        hc = xxhash.xxh32(bytes(self.board.data)).intdigest()  # should be repeatable between restarts.
        return hc

    def get_height_width(self):
        return self.board.shape[0], self.board.shape[1]

    def get_board_char_counts(self, agent_indication_embedded_value_list):
        result = {}
        unique_values = np.unique(self.board)
        for v in unique_values:
            s = np.sum(self.board == v)
            ch = self._decode_val_to_char(v, agent_indication_embedded_value_list)
            result[ch] = s.item()
        return result

    def multiworld_print(self, list_of_records, pad_length=30, agent_indication_raw_value_list=[]):
        """
        Print a number of worlds in coloured text, left to right across the screen
        @param list_of_worlds: dict e.g. [{"World":list[list[char]], "Caption":str, "Color":color code}]
        @return:
        """
        heights = [record["World"].get_height_width()[0] for record in list_of_records]
        max_lines = max(heights)
        agent_indication_embedded_value_list = [self.get_val_for_rgb_cell(cell) for cell in
                                                agent_indication_raw_value_list if cell.shape == self.raw_cell_shape]

        num_caption_lines = 1
        for world_index in range(len(list_of_records)):
            if "Caption" in list_of_records[world_index]:
                caption_list = list_of_records[world_index]["Caption"].split("\n")
                if len(caption_list) > num_caption_lines:
                    num_caption_lines = len(caption_list)

        for caption_line in range(num_caption_lines):
            line = ""
            for world_index in range(len(list_of_records)):
                caption = ""
                if "Caption" in list_of_records[world_index]:
                    caption_list = list_of_records[world_index]["Caption"].split("\n")
                    caption_list.extend(["", "", ""])
                    caption = caption_list[caption_line]
                line += caption.ljust(pad_length, " ")
            print(line)

        for line_num in range(max_lines):
            line = ""
            for world_index in range(len(list_of_records)):
                pass
                # Set the color
                color = None
                if "Color" in list_of_records[world_index]:
                    color = list_of_records[world_index]["Color"]
                    line += color
                map_line, row_length = list_of_records[world_index]["World"].get_board_line(line_num, color,
                                                                                            agent_indication_embedded_value_list)
                line += map_line
                line += nace.color_codes.color_code_black_on_white

                padding = "".join([" "] * (pad_length - row_length))
                line += padding
                # line += map_line.ljust(pad_length, " ")[len(map_line):]
            print(line)
        print()  # blank line at end

    def board_as_string(self, agent_indication_embedded_value_list):
        result = ""
        for row_number in range(self.board.shape[0]):
            for col_number in range(self.board.shape[1]):
                result += self.get_char_at_rc(row_number, col_number, agent_indication_embedded_value_list)
            result += "\n"
        return result.strip()

    def to_string_list(self):
        board_lines = []
        time_lines = []
        for row_number in range(self.board.shape[0]):
            board_line = ""
            time_line = ""
            for col_number in range(self.board.shape[1]):
                c = self.get_char_at_rc(row_number, col_number)
                board_line += c
                t = self.times[row_number][col_number]
                if t == float("-inf"):
                    time_line += "-inf,"
                else:
                    time_line += str(t) + ","
            board_lines.append(board_line)
            time_lines.append(time_line)
        return ([board_lines, (), time_lines])

    @staticmethod
    def from_string(
            board_list_of_str,
            cell_shape_rc=(45, 180, 3),
            view_dist_x=12,
            view_dist_y=5,
            agent_char_code_list=['B'],
            observed_times_str_list=None,
            numpy_seed=123,
            update_mode='ALL',
            int_2from_char_mapping=[]):
        """
        @param board_list_of_str: list[str] - each string represent 1 row.
        @return:
        """

        for line in board_list_of_str:
            if len(line.strip()) == 0:
                raise ("board lines must all be non empty and not contain space (' ')")

        add_surrounding_walls = False
        wall_thickness = 0

        char_to_rgb_mapping, rgb_to_char_mapping, raw_rgb_array = nace.numpy_utils.convert_string_list_to_cells(
            board_list_of_str,
            cell_shape_rc=cell_shape_rc,
            numpy_seed=numpy_seed)

        # lookup the rgb ndarray value of the agent
        agent_indication_raw_value_list = [char_to_rgb_mapping[agent_char_code] for agent_char_code in
                                           agent_char_code_list]

        # a map can be created with no board
        world = NPWorld(with_observed_time=True, name="from_string()",
                        initial_value=nace.world_module_numpy.UNOBSERVED_BOARD_VALUE,
                        view_dist_x=view_dist_x,
                        view_dist_y=view_dist_y,
                        rgb_to_char_mapping=rgb_to_char_mapping,
                        raw_nptype=type(list(char_to_rgb_mapping.values())[0][0][0][0]),
                        raw_cell_shape=cell_shape_rc,
                        user_set_int_2from_char_mapping=int_2from_char_mapping,
                        )

        world.multiworld_print([
            {"Caption": f"Blank Expected (no board):",
             "World": world,
             "Color": nace.color_codes.color_code_white_on_black},
        ], agent_indication_raw_value_list=agent_indication_raw_value_list)

        world.update_world_from_ground_truth_NPArray(
            raw_rgb_array,
            update_mode=update_mode,
            wall_code=nace.world_module_numpy.EMBEDDED_WALL_CODE,
            observed_at=float('-inf'),  # -inf == never seen
            agent_indication_raw_value_list=agent_indication_raw_value_list,
            cell_shape_rc=cell_shape_rc,
            add_surrounding_walls=add_surrounding_walls
        )

        if observed_times_str_list is not None:
            for row, line in enumerate(observed_times_str_list):
                time_list = line.split(",")
                for col, t_str in enumerate(time_list):
                    if t_str != '':
                        observed_time = float(t_str)
                        if row < world.times.shape[0]:
                            if col < world.times.shape[1]:
                                world.times[row + wall_thickness][col + wall_thickness] = observed_time

        world.multiworld_print([
            {"Caption": f"Should have a board:",
             "World": world,
             "Color": nace.color_codes.color_code_white_on_black},
        ], agent_indication_raw_value_list=agent_indication_raw_value_list)

        return world, agent_indication_raw_value_list

    def from_gymnasium_rgb(gym_np_world: np.ndarray,
                           agent_indication_raw_value_list=['B'],
                           add_surrounding_walls=True,
                           embedded_wall_code: int = ord('A'),
                           observed_at=float('-inf'),
                           cell_shape_rc=(60, 60),
                           store_raw_cell_data=False,
                           ):
        """

        @param gym_world: str , each \n indicates a new row, white space is stripped. agent location indicated by  agent_location_indicator_char
        @param strip_blanks:
        @param agent_location_indicator_char:  char representing the agent
        @param last_world: if not none, state and times seen copied from this world
        @param add_walls_right_and_bottom:
        @param embedded_wall_code: internal embedded value (int)
        @return:
        """

        world = NPWorld(with_observed_time=True,
                        name="from_rgb()",
                        initial_value=UNOBSERVED_BOARD_VALUE,
                        view_dist_x=3,
                        view_dist_y=2,
                        store_raw_cell_data=store_raw_cell_data)

        pre_update_agent_location_embedded_space_list, post_update_agent_location_embedded_space_list, _, _ = world.update_world_from_ground_truth_NPArray(
            gym_np_world,
            update_mode='VIEW',
            observed_at=observed_at,  # -inf == never seen
            agent_indication_raw_value_list=agent_indication_raw_value_list,
            cell_shape_rc=cell_shape_rc,
            add_surrounding_walls=add_surrounding_walls,
            wall_code=embedded_wall_code
        )
        return world, post_update_agent_location_embedded_space_list

    def get_internal_state(self):
        # do we need to store the board as well?
        int_2from_rgb_list = []
        for k in self.int_2from_rgb_mapping.keys():
            int_2from_rgb_list.append((k, self.int_2from_rgb_mapping[k]))

        int_2from_char_list = []
        for k in self.int_2from_char_mapping.keys():
            int_2from_char_list.append((k, self.int_2from_char_mapping[k]))

        result = {"int_2from_rgb_list": int_2from_rgb_list,
                  "int_2from_char_list": int_2from_char_list,
                  "raw_cell_shape": self.raw_cell_shape,
                  "raw_nptype": str(self.raw_nptype)
                  }
        # check it can be json serialised
        s = json.dumps(result)

        return result



from setuptools import setup, find_packages
import base64

install_requires = ['numba>=0.60.0',
                    'gymnasium',
                    'numba',
                    'pygame',
                    'xxhash==3.5.0',
                    'opencv-python',
                    'bidict',
                    'matplotlib', # used in testing to convert char to image
                    'getch',
                    # 'seaborn', # optional, if installed performance graphs can be generated and saved.
                    # "gymnasium[classic-control]"
                    ]


from pathlib import Path
this_directory = Path(__file__).parent

if Path(this_directory / "README_pypi.md").exists():
    import main_run_all_tests
    # we are building and or uploading the package
    long_description = (this_directory / "README_pypi.md").read_text()
    # hard coding: run all the tests
    main_run_all_tests.run_all_tests()
else:
    long_description = ""



# the .png artifacts need to be uploaded to github dwanev i.e. http://reasonedai.com/artifacts/other/



setup_kwargs = {
    'name': 'nace',
    'version': '0.0.29',
    'description': 'A re-implementation of NACE, as a pypi package, with a cleaner more general interface.', # overall description
    'long_description': long_description,
    'long_description_content_type':'text/markdown',
    'author': 'ucabdv1',
    'author_email': 'ucabdv1@ucl.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': find_packages(),
    'install_requires': install_requires,
    'python_requires': '>=3.9',
    'classifiers':[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            ],
}


setup(**setup_kwargs)


"""

cross platform, non blocking way of reading a key

https://stackoverflow.com/questions/510357/how-to-read-a-single-character-from-the-user

hmmm also does not work on MacOS....

"""
import sys,tty,os,termios
def getkey():
    old_settings = termios.tcgetattr(sys.stdin)
    tty.setcbreak(sys.stdin.fileno())
    try:
        while True:
            b = os.read(sys.stdin.fileno(), 3).decode()
            if len(b) == 3:
                k = ord(b[2])
            else:
                k = ord(b)
            key_mapping = {
                127: 'backspace',
                10: 'return',
                32: 'space',
                9: 'tab',
                27: 'esc',
                65: 'up',
                66: 'down',
                67: 'right',
                68: 'left'
            }
            return key_mapping.get(k, chr(k))
    finally:
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)


if __name__ == "__main__":
    try:
        while True:
            k = getkey()
            if k == 'esc':
                quit()
            else:
                print(k)
    except (KeyboardInterrupt, SystemExit):
        os.system('stty sane')
        print('stopping.')


from nace.nace_v3 import _build_change_sets
from nace.world_module_numpy import NPWorld
import nace


def t_1_basic():
        old_world_str_list = [
            ['oooooooooooo',
             'o          o',
             'o x        o',
             'o          o',
             'o          o',
             'o          o',
             'oooooooooooo'], (),
            ['0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,']]

        old_world, _1 = NPWorld.from_string(
            old_world_str_list[0],
            view_dist_x=12,
            view_dist_y=5,
            agent_char_code_list=['x'],
            observed_times_str_list=old_world_str_list[2]
        )

        new_world_str_list = [
            ['oooooooooooo',
             'o          o',
             'o  x       o',
             'o          o',
             'o          o',
             'o          o',
             'oooooooooooo'], (),
            ['0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,',
             '0,0,0,0,0,0,0,0,0,0,0,0,']]
        new_world, _2 = NPWorld.from_string(
            new_world_str_list[0],
            view_dist_x=12,
            view_dist_y=5,
            agent_char_code_list=['x'],
            observed_times_str_list=new_world_str_list[2])

        predicted_world, _3 = NPWorld.from_string(
            old_world_str_list[0], # NOTE: we pass in old world, i.e. we predicted no change
            view_dist_x=12,
            view_dist_y=5,
            agent_char_code_list=['x'],
            observed_times_str_list=old_world_str_list[2]
        )


        new_world.multiworld_print([
            {"World": old_world,       "Caption":"Old", "Color": nace.color_codes.color_code_white_on_blue},
            {"World": predicted_world, "Caption":"Pred", "Color": nace.color_codes.color_code_white_on_blue},
            {"World": new_world,       "Caption":"New", "Color": nace.color_codes.color_code_white_on_blue},
        ])

        action = 'move'
        focus_set = {}

        # Call the function
        new_focus_set, adjacent_change_sets, changeset0len = _build_change_sets(
            focus_set,
            old_world,
            action,
            new_world,
            predicted_world,
        )

        val = new_world.get_val_for_char('x')
        # Assertions to check the function behavior
        assert val in new_focus_set
        assert new_focus_set[val] == 1
        assert adjacent_change_sets == [{(2, 2), (2, 3)}]


def t2_diagonal_down_right():
    old_world_str_list = [
        ['oooooooooooo',
         'o          o',
         'o x        o',
         'o          o',
         'o          o',
         'o          o',
         'oooooooooooo'], (),
        ['0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,']]

    old_world, _1 = NPWorld.from_string(
        old_world_str_list[0],
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=['x'],
        observed_times_str_list=old_world_str_list[2]
    )

    new_world_str_list = [
        ['oooooooooooo',
         'o          o',
         'o          o',
         'o  x       o',
         'o          o',
         'o          o',
         'oooooooooooo'], (),
        ['0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,',
         '0,0,0,0,0,0,0,0,0,0,0,0,']]
    new_world, _2 = NPWorld.from_string(
        new_world_str_list[0],
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=['x'],
        observed_times_str_list=new_world_str_list[2])

    predicted_world, _3 = NPWorld.from_string(
        old_world_str_list[0],  # NOTE: we pass in old world, i.e. we predicted no change
        view_dist_x=12,
        view_dist_y=5,
        agent_char_code_list=['x'],
        observed_times_str_list=old_world_str_list[2]
    )

    new_world.multiworld_print([
        {"World": old_world, "Caption": "Old", "Color": color_codes.color_code_white_on_blue},
        {"World": predicted_world, "Caption": "Pred", "Color": color_codes.color_code_white_on_blue},
        {"World": new_world, "Caption": "New", "Color": color_codes.color_code_white_on_blue},
    ])

    action = 'move'
    focus_set = {}

    # Call the function
    new_focus_set, adjacent_change_sets, changeset0len = _build_change_sets(
        focus_set,
        old_world,
        action,
        new_world,
        predicted_world,
    )

    val = new_world.get_val_for_char('x')
    # Assertions to check the function behavior
    assert val in new_focus_set
    assert new_focus_set[val] == 1
    expected_change_set = [{(2, 2), (3, 3)}]
    assert len(adjacent_change_sets) == len(expected_change_set)
    for i in range(len(expected_change_set)):
        assert adjacent_change_sets[i] == expected_change_set[i]


if __name__ == '__main__':
    t_1_basic() # passes
    t2_diagonal_down_right()





"""
Handy routines to manipulate a ndarray.

"""
import os
import numpy as np
import tempfile
import matplotlib.pyplot as plt

try:
    import cv2 # pip install opencv-python
    is_cv2_installed = True
except ImportError as e:
    is_cv2_installed = False

def convert_chr_to_rgb(ch:str, rgb_shape:tuple):
    decimal_ascii = ord(ch)
    fig, ax = plt.subplots()
    fig2 = plt.figure(figsize=(.6, .15), dpi=300)
    plt.text(0.05, 0.05, ch)
    filename = str(decimal_ascii).rjust(3,"0")+".png"
    full_filename = os.path.join(tempfile.gettempdir(), filename)
    plt.savefig(full_filename)
    plt.close(fig2)
    plt.close()


    if is_cv2_installed:
        images_cv = cv2.imread(full_filename)
        rgb = np.array(images_cv, dtype=np.int64)
        assert rgb.shape == rgb_shape

    return rgb


def convert_string_list_to_cells(string_list:list[str], cell_shape_rc=(45, 180, 3), numpy_seed=None):
    char_to_rgb_mapping = {}
    rgb_to_char_mapping = {}
    rgb_shape = cell_shape_rc
    height=len(string_list)
    width=len(string_list[0])
    for i in range(height):
        assert width==len(string_list[i]) # all lines should be the same length
    raw_rgb_array = np.zeros((height * cell_shape_rc[0], width * cell_shape_rc[1], cell_shape_rc[2]), dtype=np.int64)
    if numpy_seed is not None:
        np.random.seed(numpy_seed)

    for r,line in enumerate(string_list):
        for c, ch in enumerate(line):

            if ch not in char_to_rgb_mapping:
                rgb_cell = convert_chr_to_rgb(ch,rgb_shape)
                # rgb_cell = np.random.randint(255, size=rgb_shape)
                char_to_rgb_mapping[ch] = rgb_cell
                rgb_to_char_mapping[rgb_cell.data.tobytes()] = ch

            patch = char_to_rgb_mapping[ch]
            # overwite patch
            raw_rgb_array[r * cell_shape_rc[0]:(r + 1) * cell_shape_rc[0], c * cell_shape_rc[1]:(c + 1) * cell_shape_rc[1], 0:cell_shape_rc[2]] = patch

    return char_to_rgb_mapping, rgb_to_char_mapping, raw_rgb_array


def convert_char_list_of_lists_to_cells(char_list_of_lists:list, cell_shape_rc=(60, 60, 3), numpy_seed=None ):
    # [['o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o'],
    #  ['o', ' ', ' ', ' ', 'o', ' ', ' ', ' ', 'f', ' ', ' ', 'o'],
    #  ['o', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'o'],
    #  ['o', ' ', ' ', ' ', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o'],
    #  ['o', ' ', 'x', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'o'],
    #  ['o', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'u', ' ', ' ', 'o'],
    #  ['o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o', 'o']]
    char_to_rgb_mapping = {}
    rgb_to_char_mapping = {}
    rgb_shape = cell_shape_rc
    height=len(char_list_of_lists)
    width=len(char_list_of_lists[0])
    raw_rgb_array = np.zeros((height * cell_shape_rc[0], width * cell_shape_rc[1], cell_shape_rc[2]), dtype=np.int64)
    if numpy_seed is not None:
        np.random.seed(numpy_seed)

    for r,line_list in enumerate(char_list_of_lists):
        for c, ch in enumerate(line_list):
            if ch not in char_to_rgb_mapping:
                rgb_cell = np.random.randint(255, size=rgb_shape)
                char_to_rgb_mapping[ch] = rgb_cell
                rgb_to_char_mapping[rgb_cell.data.tobytes()] = ch

            patch = char_to_rgb_mapping[ch]
            # overwite patch
            raw_rgb_array[r * cell_shape_rc[0]:(r + 1) * cell_shape_rc[0], c * cell_shape_rc[1]:(c + 1) * cell_shape_rc[1], 0:cell_shape_rc[2]] = patch

    return char_to_rgb_mapping, rgb_to_char_mapping, raw_rgb_array